import os
import tempfile
import pandas as pd
import csv
from patsy.highlevel import dmatrices
from nidm.experiment.Query import GetProjectsUUID
import click
from nidm.experiment.tools.click_base import cli
from nidm.experiment.tools.rest import RestParser
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.metrics import silhouette_score, adjusted_rand_score
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder, MinMaxScaler
from sklearn.cluster import AffinityPropagation
from sklearn import metrics
from sklearn import preprocessing

@cli.command()
@click.option("--nidm_file_list", "-nl", required=True,
              help="A comma separated list of NIDM files with full path")
@click.option("-variables", required=False,
                 help="This parameter is for the variables the user would like to complete the k-means algorithm on.\nThe way this looks in the command is python3 nidm_kmeans.py -nl MTdemog_aseg_v2.ttl -v \"fs_003343,age*sex,sex,age,group,age*group,bmi\"")
@click.option("--output_file", "-o", required=False,
              help="Optional output file (TXT) to store results of the linear regression, contrast, and regularization")
def full_ap(nidm_file_list, output_file, variables):
    global v  # Needed to do this because the code only used the parameters in the first method, meaning I had to move it all to method 1.
    v = variables.strip()  # used in data_aggregation, linreg(), spaces stripped from left and right
    global o  # used in dataparsing()
    o = output_file
    global n  # used in data_aggregation()
    n = nidm_file_list
    data_aggregation()
    dataparsing()
    ap()


def data_aggregation():  # all data from all the files is collected
    """    This function provides query support for NIDM graphs.   """
    # query result list
    results = []
    # if there is a CDE file list, seed the CDE cache
    if v:  # ex: fs_00343 ~ age + sex + group
        print("***********************************************************************************************************")
        command = "python nidm_kmeans.py -nl " + n + " -variables \"" + v + "\" "

        print("Your command was: " + command)
        if (o is not None):
            f = open(o, "w")
            f.write("Your command was " + command)
            f.close()
        verbosity = 0
        restParser = RestParser(verbosity_level=int(verbosity))
        restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
        global df_list  # used in dataparsing()
        df_list = []
        # set up uri to do fields query for each nidm file
        global file_list
        file_list = n.split(",")
        df_list_holder = {}
        for i in range(len(file_list)):
            df_list_holder[i] = []
        df_holder = {}
        for i in range(len(file_list)):
            df_holder[i] = []
        global condensed_data_holder
        condensed_data_holder = {}
        for i in range(len(file_list)):
            condensed_data_holder[i] = []
        count = 0
        not_found_count = 0
        for nidm_file in file_list:
            # get project UUID
            project = GetProjectsUUID([nidm_file])
            # split the model into its constituent variables
            global full_model_variable_list
            full_model_variable_list=[]
            global model_list
            model_list = v.split(",")
            for i in range(len(model_list)):  # here, we remove any leading or trailing spaces
                model_list[i] = model_list[i].strip()
            global vars  # used in dataparsing()
            vars = ""
            for i in range(len(model_list) - 1, -1, -1):
                full_model_variable_list.append(model_list[i])  # will be used in the regularization, but we need the full list
                if "*" in model_list[i]:  # removing the star term from the columns we're about to pull from data
                    model_list.pop(i)
                else:
                    vars = vars + model_list[i] + ","
            vars = vars[0:len(vars) - 1]
            uri = "/projects/" + project[0].toPython().split("/")[-1] + "?fields=" + vars
            # get fields output from each file and concatenate
            df_list_holder[count].append(pd.DataFrame(restParser.run([nidm_file], uri)))
            df = pd.concat(df_list_holder[count])
            with tempfile.NamedTemporaryFile(delete=False) as temp:  # turns the dataframe into a temporary csv
                df.to_csv(temp.name + '.csv')
                temp.close()
            data = list(csv.reader(open(temp.name + '.csv')))  # makes the csv a 2D list to make it easier to call the contents of certain cells
            numcols = (len(data) - 1) // (len(model_list))  # Finds the number of columns in the original dataframe
            global condensed_data  # also used in linreg()
            condensed_data_holder[count] = [[0] * (len(model_list))]  # makes an array 1 row by the number of necessary columns
            for i in range(numcols):  # makes the 2D array big enough to store all of the necessary values in the edited dataset
                condensed_data_holder[count].append([0] * (len(model_list)))
            for i in range(len(model_list)):  # stores the independent variable names in the first row
                condensed_data_holder[count][0][i] = model_list[i]
            numrows = 1  # begins at the first row to add data
            fieldcolumn = 0  # the column the variable name is in in the original dataset
            valuecolumn = 0  # the column the value is in in the original dataset
            datacolumn = 0  # if it is identified by the dataElement name instead of the field's name
            not_found_list = []
            for i in range(len(data[0])):
                if data[0][i] == 'sourceVariable':  # finds the column where the variable names are
                    fieldcolumn = i
                elif data[0][i] == 'source_variable':  # finds the column where the variable names are
                    fieldcolumn = i
                elif data[0][i] == 'isAbout':
                    aboutcolumn = i
                elif data[0][i] == 'label':
                    namecolumn = i  # finds the column where the variable names are
                elif data[0][i] == 'value':
                    valuecolumn = i  # finds the column where the values are
                elif data[0][i] == 'dataElement':  # finds the column where the data element is if necessary
                    datacolumn = i
            for i in range(
                    len(condensed_data_holder[count][
                            0])):  # starts iterating through the dataset, looking for the name in that
                for j in range(1, len(data)):  # column, so it can append the values under the proper variables
                    try:
                        split_url = condensed_data_holder[count][0][i].split("/")
                        if data[j][fieldcolumn] == condensed_data_holder[count][0][
                            i]:  # in the dataframe, the name is in column 3
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif data[j][aboutcolumn] == condensed_data_holder[count][0][
                            i]:
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif data[j][aboutcolumn] == split_url[
                            len(split_url) - 1]:  # this is in case the uri only works by querying the part after the last backslash
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values

                        elif data[j][namecolumn] == condensed_data_holder[count][0][
                            i]:  # in the dataframe, the name is in column 12
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif condensed_data_holder[count][0][i] == data[j][
                            datacolumn]:  # in the dataframe, the name is in column 9
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                    except IndexError:
                        numrows = numrows + 1
                numrows = 1  # resets to the first row for the next variable
            temp_list = condensed_data_holder[count]
            for j in range(len(temp_list[0]) - 1, 0,-1):  # if the software appends a column with 0 as the heading, it removes this null column
                if temp_list[0][j] == "0" or temp_list[0][j] == "NaN":
                    for row in condensed_data_holder[count]:
                        row.pop(j)
            rowsize = len(condensed_data_holder[count][0])
            count1 = 0
            for i in range(0, rowsize):
                for row in condensed_data_holder[count]:
                    if row[i] == 0 or row[i] == "NaN" or row[i] == "0":
                        count1 = count1 + 1
                if count1 > len(condensed_data_holder[count]) - 2:
                    not_found_list.append(condensed_data_holder[count][0][i])
                count1 = 0
            for i in range(len(condensed_data_holder[count][0])):
                if " " in condensed_data_holder[count][0][i]:
                    condensed_data_holder[count][0][i] = condensed_data_holder[count][0][i].replace(" ", "_")
            for i in range(len(vars)):
                if " " in vars[i]:
                    vars[i] = vars[i].replace(" ", "_")
            count = count + 1
            if len(not_found_list) > 0:
                print(
                        "***********************************************************************************************************")
                print()
                print("Your variables were " + v)
                print()
                print(
                        "The following variables were not found in " + nidm_file + ". The model cannot run because this will skew the data. Try checking your spelling or use nidm_query.py to see other possible variables.")
                if (o is not None):
                    f = open(o, "a")
                    f.write("Your variables were " + v)
                    f.write(
                            "The following variables were not found in " + nidm_file + ". The model cannot run because this will skew the data. Try checking your spelling or use nidm_query.py to see other possible variables.")
                    f.close()
                for i in range(0, len(not_found_list)):
                    print(str(i + 1) + ". " + not_found_list[i])
                    if (o is not None):
                        f = open(o, "a")
                        f.write(str(i + 1) + ". " + not_found_list[i])
                        f.close()
                for j in range(len(not_found_list) - 1, 0, -1):
                    not_found_list.pop(j)
                not_found_count = not_found_count + 1
                print()
        if not_found_count > 0:
            exit(1)


    else:
        print("ERROR: No query parameter provided.  See help:")
        print()
        os.system("pynidm query --help")
        exit(1)

def dataparsing(): #The data is changed to a format that is usable by the linear regression method
    global condensed_data
    condensed_data = []
    for i in range(0, len(file_list)):
        condensed_data = condensed_data + condensed_data_holder[i]
    x = pd.read_csv(opencsv(condensed_data))  # changes the dataframe to a csv to make it easier to work with
    x.head()  # prints what the csv looks like
    x.dtypes  # checks data format
    obj_df = x.select_dtypes  # puts all the variables in a dataset
    x.shape  # says number of rows and columns in form of tuple
    x.describe()  # says dataset statistics
    obj_df = x.select_dtypes(
        include=['object']).copy()  # takes everything that is an object (not float or int) and puts it in a new dataset
    obj_df.head()  # prints the new dataset
    int_df = x.select_dtypes(include=['int64']).copy()  # takes everything that is an int and puts it in a new dataset
    float_df = x.select_dtypes(
        include=['float64']).copy()  # takes everything that is a float and puts it in a new dataset
    df_int_float = pd.concat([float_df, int_df], axis=1)
    stringvars = []  # starts a list that will store all variables that are not numbers
    for i in range(1, len(condensed_data)):  # goes through each variable
        for j in range(len(condensed_data[0])):  # in the 2D array
            try:  # if the value of the field can be turned into a float (is numerical)
                float(condensed_data[i][j])  # this means it's a number
            except ValueError:  # if it can't be (is a string)
                if condensed_data[0][
                    j] not in stringvars:  # adds the variable name to the list if it isn't there already
                    stringvars.append(condensed_data[0][j])
    le = preprocessing.LabelEncoder()  # anything involving le shows the encoding of categorical variables
    for i in range(len(stringvars)):
        le.fit(obj_df[stringvars[i]].astype(str))
    obj_df_trf = obj_df.astype(str).apply(le.fit_transform)  # transforms the categorical variables into numbers.
    global df_final  # also used in linreg()
    if not obj_df_trf.empty:
        df_final = pd.concat([df_int_float, obj_df_trf], axis=1)  # join_axes=[df_int_float.index])
    else:
        df_final = df_int_float
    df_final.head()  # shows the final dataset with all the encoding
    print(df_final)  # prints the final dataset
    print()
    print("***********************************************************************************************************")
    print()
    if (o is not None):
        f = open(o, "a")
        f.write(df_final.to_string(header=True, index=True))
        f.write(
            "\n\n***********************************************************************************************************")
        f.write("\n\nModel Results: ")
        f.close()
def ap():
    index = 0
    global levels  # also used in contrasting()
    levels = []
    for i in range(1, len(condensed_data)):
        if condensed_data[i][index] not in levels:
            levels.append(condensed_data[i][index])
    for i in range(len(levels)):
        levels[i] = i

    # Beginning of the linear regression
    global X
    #global y
    #Unsure on how to procede here with interacting variables, since I'm sure dmatrices won't work

    scaler = MinMaxScaler()

    for i in range(len(model_list)):
        scaler.fit(df_final[[model_list[i]]])
        df_final[[model_list[i]]] = scaler.transform(df_final[[model_list[i]]])

    X = df_final[model_list]

    af = AffinityPropagation(preference=-50).fit(X)
    cluster_center_indices=af.cluster_centers_indices_
    labels = af.labels_
    n_clusters_ = len(cluster_center_indices)

    print('Estimated number of clusters: %d' % n_clusters_)
    #print("Homogeneity: %0.3f" % metrics.homogeneity_score(labels_true, labels))
    #print("Completeness: %0.3f" % metrics.completeness_score(labels_true, labels))
    #print("V-measure: %0.3f" % metrics.v_measure_score(labels_true, labels))
    #print("Adjusted Rand Index: %0.3f" % metrics.adjusted_rand_score(labels_true, labels))
    #print("Adjusted Mutual Information: %0.3f" % metrics.adjusted_mutual_info_score(labels_true, labels))
    print("Silhouette Coefficient: %0.3f" % metrics.silhouette_score(X, labels, metric='sqeuclidean'))

    sns.scatterplot(data=X, x=model_list[0], y=model_list[1], hue=af, palette = "gnuplot")
    plt.xlabel(model_list[1])
    plt.ylabel(model_list[0])
    title = "Clustering results of "
    for i in range(len(model_list)):
        title = title + model_list[i] + ","
    title = title[0:len(title)-1]
    plt.title(title)
    plt.show()
    if (o is not None):
        f = open(o, "a")
        f.close()
def opencsv(data):
    """saves a list of lists as a csv and opens"""
    import tempfile
    import os
    import csv
    handle, fn = tempfile.mkstemp(suffix='.csv')
    with os.fdopen(handle,"w", encoding='utf8',errors='surrogateescape',newline='') as f:
        writer = csv.writer(f)
        writer.writerows(data)
    return fn

# it can be used calling the script `python nidm_query.py -nl ... -q ..
if __name__ == "__main__":
    full_ap()



import urllib
import re
import sys

import pytest
import rdflib

from nidm.experiment import Project, Session, AssessmentAcquisition, AssessmentObject, Acquisition, AcquisitionObject, Query
from nidm.core import Constants
from nidm.experiment.tools.rest import RestParser
from nidm.experiment.tools.tests.test_rest_statistics import BRAIN_VOL_FILES

from nidm.experiment.tools.nidm_linreg import linear_regression

import os
from os.path import join,sep
from pathlib import Path
from rdflib import Graph, util, URIRef
import json
from io import TextIOWrapper, BytesIO
import subprocess
from subprocess import PIPE
import tempfile
import click

@pytest.fixture(scope="module", autouse="True")
def setup():

    for f in ['./cmu_a.nidm.ttl', 'caltech.nidm.ttl']:
        if Path(f).is_file():
            os.remove(f)

    if not Path('./cmu_a.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
            "cmu_a.nidm.ttl"
        )

    if not Path('./caltech.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/Caltech/nidm.ttl",
            "caltech.nidm.ttl"
        )

def call_click_command(cmd, *args, **kwargs):
    """ Wrapper to call a click command

    :param cmd: click cli command function to call
    :param args: arguments to pass to the function
    :param kwargs: keywrod arguments to pass to the function
    :return: None
    """

    # Get positional arguments from args
    arg_values = {c.name: a for a, c in zip(args, cmd.params)}
    args_needed = {c.name: c for c in cmd.params
                   if c.name not in arg_values}

    # build and check opts list from kwargs
    opts = {a.name: a for a in cmd.params if isinstance(a, click.Option)}
    for name in kwargs:
        if name in opts:
            arg_values[name] = kwargs[name]
        else:
            if name in args_needed:
                arg_values[name] = kwargs[name]
                del args_needed[name]
            else:
                raise click.BadParameter(
                    "Unknown keyword argument '{}'".format(name))


    # check positional arguments list
    for arg in (a for a in cmd.params if isinstance(a, click.Argument)):
        if arg.name not in arg_values:
            raise click.BadParameter("Missing required positional"
                                     "parameter '{}'".format(arg.name))

    # build parameter lists
    opts_list = sum(
        [[o.opts[0], arg_values[n]] for n, o in opts.items()], [])
    args_list = [str(v) for n, v in arg_values.items() if n not in opts]

    # call the command
    try:
        cmd(opts_list + args_list)
    except:
        pass



def test_simple_model():


    # run linear regression tool with simple model and evaluate output
    dirname = os.path.dirname(__file__)
    linreg_dirname = join(sep+join(*(dirname.split(sep)[:-1])))


    arguments = (dict(nidm_file_list =",".join(BRAIN_VOL_FILES),
                      ml='fs_000008 = DX_GROUP + http://uri.interlex.org/ilx_0100400',
                      ctr=None,regularization=None,output_file="output.txt"))

    call_click_command(linear_regression,*arguments,**arguments)
   

    if os.path.exists("output.txt"):
        fp = open("output.txt", "r")
        out = fp.read()
        fp.close()
        os.remove("output.txt")



    # check if model was read correctly
    assert "fs_000008 ~ ilx_0100400 + DX_GROUP" in out

    # check correct number of observations
    assert "No. Observations:                  53" in out

    # check model coefficients
    assert "const          27.7816      4.378      6.345      0.000      18.988      36.576" in out
    assert "ilx_0100400    -0.1832      0.173     -1.061      0.294      -0.530       0.164" in out
    assert "DX_GROUP        3.4908      4.031      0.866      0.391      -4.605      11.587" in out

def test_model_with_contrasts():
    # run linear regression tool with simple model and evaluate output

    arguments = (dict(nidm_file_list=",".join(BRAIN_VOL_FILES),
                      ml='fs_000008 = DX_GROUP + http://uri.interlex.org/ilx_0100400',
                      ctr="DX_GROUP", regularization=None, output_file="output.txt"))

    call_click_command(linear_regression, *arguments, **arguments)


    if os.path.exists('output.txt'):
        fp = open('output.txt', "r")
        out = fp.read()
        fp.close()
        os.remove('output.txt')

    # print(out)

    # check if model was read correctly
    assert "fs_000008 ~ ilx_0100400 + DX_GROUP" in out

    # check correct number of observations
    assert "No. Observations:                  53" in out

    # check model coefficients for different codings
    assert "C(DX_GROUP, Treatment)[T.1]     0.7307      4.209      0.174      0.863      -7.727       9.189" in out
    assert "C(DX_GROUP, Treatment)[T.2]    32.7462     15.984      2.049      0.046       0.625      64.868" in out
    assert "C(DX_GROUP, Simple)[Simp.0]     0.7307      4.209      0.174      0.863      -7.727       9.189" in out
    assert "C(DX_GROUP, Simple)[Simp.1]    32.7462     15.984      2.049      0.046       0.625      64.868" in out
    assert "C(DX_GROUP, Sum)[S.0]   -11.1590      5.713     -1.953      0.057     -22.639       0.321" in out
    assert "C(DX_GROUP, Sum)[S.1]   -10.4283      5.631     -1.852      0.070     -21.743       0.887" in out
    assert "C(DX_GROUP, Diff)[D.0]     0.7307      4.209      0.174      0.863      -7.727       9.189" in out
    assert "C(DX_GROUP, Diff)[D.1]    32.0155     15.896      2.014      0.050       0.070      63.961" in out
    assert "C(DX_GROUP, Helmert)[H.1]     0.3653      2.104      0.174      0.863      -3.864       4.594" in out
    assert "C(DX_GROUP, Helmert)[H.2]    10.7936      5.267      2.049      0.046       0.209      21.378" in out

@pytest.mark.skip(reason="regularization weights seem to be different depending on the platform")
def test_model_with_contrasts_reg_L1():


    arguments = (dict(nidm_file_list=",".join(BRAIN_VOL_FILES),
                      ml='fs_000008 = DX_GROUP + http://uri.interlex.org/ilx_0100400',
                      ctr="DX_GROUP", regularization="L1", output_file="output.txt"))

    call_click_command(linear_regression, *arguments, **arguments)

    if os.path.exists('output.txt'):
        fp = open('output.txt', "r")
        out = fp.read()
        fp.close()
        os.remove('output.txt')

    print(out)

    # check if model was read correctly
    assert "fs_000008 ~ ilx_0100400 + DX_GROUP" in out

    # check correct number of observations
    assert "No. Observations:                  53" in out

    # check lasso regression
    assert "Alpha with maximum likelihood (range: 1 to 700) = 43.000000" in out
    assert "Current Model Score = 0.000000" in out
    assert "ilx_0100400 	 -0.000000" in out
    assert "DX_GROUP 	 0.000000" in out
    assert "Intercept: 26.000000" in out

@pytest.mark.skip(reason="regularization weights seem to be different depending on the platform")
def test_model_with_contrasts_reg_L2():

    arguments = (dict(nidm_file_list=",".join(BRAIN_VOL_FILES),
                      ml='fs_000008 = DX_GROUP + http://uri.interlex.org/ilx_0100400',
                      ctr="DX_GROUP", regularization="L2", output_file="output.txt"))

    call_click_command(linear_regression, *arguments, **arguments)

    if os.path.exists('output.txt'):
        fp = open('output.txt', "r")
        out = fp.read()
        fp.close()
        os.remove('output.txt')

    # print(out)

    # check if model was read correctly
    assert "fs_000008 ~ ilx_0100400 + DX_GROUP" in out

    # check correct number of observations
    assert "No. Observations:                  53" in out

    # check lasso regression
    assert "Alpha with maximum likelihood (range: 1 to 700) = 699.000000" in out
    assert "Current Model Score = 0.017618" in out
    assert "ilx_0100400 	 -0.148397" in out
    assert "DX_GROUP 	 0.071356" in out
    assert "Intercept: 28.951297" in out


# coding=utf-8
# !/usr/bin/env python

# *******************************************************************************************************
# *******************************************************************************************************
#  nidm_linreg.py
#  License: Apache License, Version 2.0
# *******************************************************************************************************
# *******************************************************************************************************
# Date: 10-09-21                 Coded by: Ashmita Kumar (ashmita.kumar@gmail.com)
# Filename: nidm_linreg.py
#
# Program description:  This program provides a tool to complete a linear regression on nidm files
#
#
# *******************************************************************************************************
# Development environment: Python - PyCharm IDE
#
# *******************************************************************************************************
# System requirements:  Python 3.X
# Libraries: os, sys, tempfile, pandas, click, nidm, csv, sklearn, numpy, statsmodel.api, patsy.contrasts
# *******************************************************************************************************
# Start date: 6-15-20
# Update history:
# DATE            MODIFICATION				Who
#
#
# *******************************************************************************************************
# Programmer comments:
#
#
# *******************************************************************************************************
# *******************************************************************************************************
import os
import sys
from os import system
import tempfile
import pandas as pd
import csv
from patsy.highlevel import dmatrices
from nidm.experiment.Query import GetProjectsUUID
import click
from nidm.experiment.tools.click_base import cli
from nidm.experiment.tools.rest import RestParser
import numpy as np
try:
    from sklearn.linear_model import LinearRegression
    from sklearn import preprocessing
    from sklearn.linear_model import Ridge
    from sklearn.linear_model import Lasso
    from sklearn.model_selection import cross_val_score
except:
    system('python -m pip install --upgrade pip sklearn')
    from sklearn.linear_model import LinearRegression
    from sklearn import preprocessing
    from sklearn.linear_model import Ridge
    from sklearn.linear_model import Lasso
    from sklearn.model_selection import cross_val_score

try:
    import statsmodels.api as sm
    from statsmodels.formula.api import ols
except:
    system('python -m pip install --upgrade pip statsmodels')
    import statsmodels.api as sm
    from statsmodels.formula.api import ols

try:
    from statistics import mean
except:
    system('python -m pip install --upgrade pip statistics')
    from statistics import mean

try:
    from patsy.contrasts import Treatment
    from patsy.contrasts import ContrastMatrix
    from patsy.contrasts import Sum
    from patsy.contrasts import Diff
    from patsy.contrasts import Helmert
except:
    system('python -m pip install --upgrade pip patsy')
    from patsy.contrasts import Treatment
    from patsy.contrasts import ContrastMatrix
    from patsy.contrasts import Sum
    from patsy.contrasts import Diff
    from patsy.contrasts import Helmert

MAX_ALPHA = 700
#Defining the parameters of the commands.
@cli.command()
@click.option("--nidm_file_list", "-nl", required=True,
              help="A comma separated list of NIDM files with full path")
@click.option("--ctr", "-contrast", required=False,
              help="This parameter will show differences in relationship by group (e.g. -contrast age*sex,group). It can be one variable, interacting variables, or multiple")
@click.option("--ml", "-model", required=True,
                 help="This parameter will return the results of the linear regression from all nidm files supplied\nThe way this looks in the command is python3 nidm_linreg.py -nl MTdemog_aseg_v2.ttl -model \"fs_003343 = age*sex + sex + age + group + age*group + bmi\" -contrast group -r L1")
@click.option("--output_file", "-o", required=False,
              help="Optional output file (TXT) to store results of the linear regression, contrast, and regularization")
@click.option("--regularization", "-r", required=False,
              help="This parameter will return the results of the linear regression with L1 or L2 regularization depending on the type specified, and the weight with the maximum likelihood solution")

def linear_regression(nidm_file_list, output_file, ml, ctr, regularization):
    """
        This function provides a tool to complete a linear regression on NIDM data with optional contrast and regularization.
        """


    #NOTE: Every time I make a global variable, it is because I need it in at least one other method.
    global c #used in linreg(), contrasting()
    c = ctr #Storing all important parameters in global variables so they can be accessed in other methods
    global m #Needed to do this because the code only used the parameters in the first method, meaning I had to move it all to method 1.
    m = ml.strip() #used in data_aggregation, linreg(), spaces stripped from left and right
    global o #used in dataparsing()
    o = output_file
    global n #used in data_aggregation()
    n = nidm_file_list
    global r
    r = regularization
    data_aggregation() #collects data
    dataparsing() #converts it to proper format
    l = linreg() #performs linear regression
    contrasting() #performs contrast
    regularizing() #performs regularization

def data_aggregation(): #all data from all the files is collected
    """
            This function provides query support for NIDM graphs.
            """
    # query result list
    results = []
    # if there is a CDE file list, seed the CDE cache
    if m:  # ex: fs_00343 ~ age + sex + group
        print("***********************************************************************************************************")
        command = "pynidm linear-regression -nl " + n + " -model \"" + m + "\" "
        if c:
            command = command + "-contrast \"" + c + "\" "
        if r:
            command = command + "-r " + r + " "
        print("Your command was: " + command)
        if (o is not None):
            f = open(o, "w")
            f.write("Your command was " + command)
            f.close()
        verbosity = 0
        restParser = RestParser(verbosity_level=int(verbosity))
        restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
        global df_list #used in dataparsing()
        df_list = []
        # set up uri to do fields query for each nidm file
        global file_list
        file_list = n.split(",")
        df_list_holder = {}
        for i in range(len(file_list)):
            df_list_holder[i] = []
        df_holder = {}
        for i in range(len(file_list)):
            df_holder[i] = []
        global condensed_data_holder
        condensed_data_holder = {}
        for i in range(len(file_list)):
            condensed_data_holder[i] = []

        count = 0
        not_found_count = 0
        for nidm_file in file_list:
            # get project UUID
            project = GetProjectsUUID([nidm_file])
            # split the model into its constituent variables
            global full_model_variable_list
            #below, we edit the model so it splits by +,~, or =. However, to help it out in catching everything
            #we replaced ~ and = with a + so that we can still use split. Regex wasn't working.
            plus_replace = m
            if "~" in m:
                plus_replace = m.replace('~','+')
            elif "=" in m:
                plus_replace = m.replace('=', '+')
            elif "," in m:
                plus_replace = m.replace(",", '+')
            model_list = plus_replace.split("+")
            for i in range(len(model_list)): #here, we remove any leading or trailing spaces
                model_list[i] = model_list[i].strip()
            full_model_variable_list = []
            # set the dependent variable to the one dependent variable in the model
            global dep_var #used in dataparsing(), linreg(), and contrasting()
            dep_var = model_list[0]
            # join the independent variables into a comma-separated list to make it easier to call from the uri
            global ind_vars #used in dataparsing()
            ind_vars = ""
            for i in range(len(model_list)-1, 0, -1):
                full_model_variable_list.append(model_list[i]) #will be used in the regularization, but we need the full list
                if "*" in model_list[i]: #removing the star term from the columns we're about to pull from data
                    model_list.pop(i)
                elif model_list[i] == dep_var:
                    model_list.pop(i)
                    print("\n\nAn independent variable cannot be the same as the dependent variable. This prevents the model from running accurately.")
                    print("Please try a different model removing \"" + dep_var + "\" from either the right or the left side of the equation.\n\n")
                    if (o is not None):
                        f = open(o, "a")
                        f.write("\n\nAn independent variable cannot be the same as the dependent variable. This prevents the model from running accurately.")
                        f.write("Please try a different model removing \"" + dep_var + "\" from either the right or the left side of the equation.")
                        f.close()
                    exit(1)
                else:
                    ind_vars = ind_vars + model_list[i] + ","
            ind_vars = ind_vars[0:len(ind_vars) - 1]
            uri = "/projects/" + project[0].toPython().split("/")[-1] + "?fields=" + ind_vars + "," + dep_var
            # get fields output from each file and concatenate
            df_list_holder[count].append(pd.DataFrame(restParser.run([nidm_file], uri)))
            #global dep_var
            df = pd.concat(df_list_holder[count])
            with tempfile.NamedTemporaryFile(delete=False) as temp:  # turns the dataframe into a temporary csv
                df.to_csv(temp.name + '.csv')
                temp.close()
            data = list(csv.reader(open(
                temp.name + '.csv')))  # makes the csv a 2D list to make it easier to call the contents of certain cells

            global independentvariables  # used in linreg
            independentvariables = ind_vars.split(",")  # makes a list of the independent variables
            numcols = (len(data) - 1) // (
                        len(independentvariables) + 1)  # Finds the number of columns in the original dataframe
            global condensed_data  # also used in linreg()
            condensed_data_holder[count] = [
                [0] * (len(independentvariables) + 1)]  # makes an array 1 row by the number of necessary columns
            for i in range(
                    numcols):  # makes the 2D array big enough to store all of the necessary values in the edited dataset
                condensed_data_holder[count].append([0] * (len(independentvariables) + 1))
            for i in range(len(independentvariables)):  # stores the independent variable names in the first row
                condensed_data_holder[count][0][i] = independentvariables[i]
            condensed_data_holder[count][0][-1] = str(dep_var)  # stores the dependent variable name in the first row
            numrows = 1  # begins at the first row to add data
            fieldcolumn = 0  # the column the variable name is in in the original dataset
            valuecolumn = 0  # the column the value is in in the original dataset
            datacolumn = 0  # if it is identified by the dataElement name instead of the field's name
            not_found_list = []
            for i in range(len(data[0])):
                if data[0][i] == 'sourceVariable':  # finds the column where the variable names are
                    fieldcolumn = i
                elif data[0][i] == 'source_variable':  # finds the column where the variable names are
                    fieldcolumn = i
                elif data[0][i] == 'isAbout':
                    aboutcolumn = i
                elif data[0][i] == 'label':
                    namecolumn = i  # finds the column where the variable names are
                elif data[0][i] == 'value':
                    valuecolumn = i  # finds the column where the values are
                elif data[0][i] == 'dataElement':  # finds the column where the data element is if necessary
                    datacolumn = i
            for i in range(
                    len(condensed_data_holder[count][0])):  # starts iterating through the dataset, looking for the name in that
                for j in range(1, len(data)):  # column, so it can append the values under the proper variables
                    try:
                        split_url = condensed_data_holder[count][0][i].split("/")
                        for k in range(0, len(full_model_variable_list)):
                            if "/" in full_model_variable_list[k]:
                                full_model_variable_list[k] = split_url[len(split_url) - 1]
                        if data[j][fieldcolumn] == condensed_data_holder[count][0][i]:  # in the dataframe, the name is in column 3
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif data[j][aboutcolumn] == condensed_data_holder[count][0][
                            i]:
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif data[j][aboutcolumn] == split_url[len(split_url)-1]: #this is in case the uri only works by querying the part after the last backslash
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif condensed_data_holder[count][0][
                            i] in data[j][aboutcolumn]: #this is in case the uri only works by querying the part after the last backslash
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif data[j][namecolumn] == condensed_data_holder[count][0][i]:  # in the dataframe, the name is in column 12
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                        elif condensed_data_holder[count][0][i] == data[j][datacolumn]:  # in the dataframe, the name is in column 9
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn]  # in the dataframe, the value is in column 2
                            numrows = numrows + 1  # moves on to the next row to add the proper values
                    except IndexError:
                        numrows = numrows + 1
                numrows = 1  # resets to the first row for the next variable
            temp_list = condensed_data_holder[count]
            for j in range(len(temp_list[0])-1, 0,-1):  # if the software appends a column with 0 as the heading, it removes this null column
                if temp_list[0][j] == "0" or temp_list[0][j] == "NaN":
                    for row in condensed_data_holder[count]:
                        row.pop(j)
            rowsize = len(condensed_data_holder[count][0])
            count1 = 0
            for i in range(0, rowsize):
                for row in condensed_data_holder[count]:
                    if row[i] == 0 or row[i] == "NaN" or row[i] == "0":
                        count1 = count1 + 1
                if count1 > len(condensed_data_holder[count]) - 2:
                    not_found_list.append(condensed_data_holder[count][0][i])
                count1 = 0
            for i in range(len(condensed_data_holder[count][0])):
                if " " in condensed_data_holder[count][0][i]:
                    condensed_data_holder[count][0][i] = condensed_data_holder[count][0][i].replace(" ", "_")
            for i in range(len(independentvariables)):
                if "/" in independentvariables[i]:
                    splitted = independentvariables[i].split("/")
                    independentvariables[i] = splitted[len(splitted)-1]
                if " " in independentvariables[i]:
                    independentvariables[i] = independentvariables[i].replace(" ", "_")
            if " " in dep_var:
                dep_var = dep_var.replace(" ", "_")
            count = count + 1
            if len(not_found_list) > 0:
                print(
                    "***********************************************************************************************************")
                print()
                print("Your model was " + m)
                print()
                print(
                    "The following variables were not found in " + nidm_file + ". The model cannot run because this will skew the data. Try checking your spelling or use nidm_query.py to see other possible variables.")
                if (o is not None):
                    f = open(o, "a")
                    f.write("Your model was " + m)
                    f.write(
                        "The following variables were not found in " + nidm_file + ". The model cannot run because this will skew the data. Try checking your spelling or use nidm_query.py to see other possible variables.")
                    f.close()
                for i in range(0, len(not_found_list)):
                    print(str(i + 1) + ". " + not_found_list[i])
                    if (o is not None):
                        f = open(o, "a")
                        f.write(str(i + 1) + ". " + not_found_list[i])
                        f.close()
                for j in range(len(not_found_list)-1, 0,-1):
                    not_found_list.pop(j)
                not_found_count = not_found_count + 1
                print()
        if not_found_count > 0:
            exit(1)


    else:
        print("ERROR: No query parameter provided.  See help:")
        print()
        os.system("pynidm linreg --help")
        exit(1)

def dataparsing(): #The data is changed to a format that is usable by the linear regression method
    global condensed_data
    condensed_data = []
    for i in range(0, len(file_list)):
        condensed_data = condensed_data + condensed_data_holder[i]
    for i in range(len(condensed_data[0])):
        if "/" in condensed_data[0][i]: #change any URLs to just the last part so contrasting works.
            splitted = condensed_data[0][i].split("/")
            condensed_data[0][i] = splitted[len(splitted) - 1]

    """In this section, if there are less than 20 points, the model will be innacurate and there are too few variables for regularization.
    That means that we warn the user that such errors can occur and ask them if they want to proceed.
    The answer is stored in answer. If the user responds with N, it exits the code after writing the error to the output file (if there is one).
    If the user says Y instead, the code runs, but stops before doing the regularization."""
    global answer
    answer = "?"
    if(len(condensed_data)-1)<20:
        print("\nYour data set has less than 20 points, which means the model calculated may not be accurate due to a lack of data. ")
        print("This means you cannot regularize the data either.")
        import warnings
        warnings.filterwarnings("ignore")
        answer = input("Continue anyways? Y or N: ")
        if (o is not None):
            f = open(o, "a")
            f.write("Your model was " + m)
            f.write(
                "\n\nThere was a lack of data (<20 points) in your model, which may result in inaccuracies. In addition, a regularization cannot and will not be performed.\n")
            f.close()
    if "n" in answer.lower():
        print("\nModel halted.")
        if (o is not None):
            f = open(o, "a")
            f.write("Your model was " + m)
            f.write("Due to a lack of data (<20 points), you stopped the model because the results may have been innacurate.")
            f.close()
        exit(1)
    x = pd.read_csv(opencsv(condensed_data))  # changes the dataframe to a csv to make it easier to work with
    x.head()  # prints what the csv looks like
    x.dtypes  # checks data format
    obj_df = x.select_dtypes  # puts all the variables in a dataset
    x.shape  # says number of rows and columns in form of tuple
    x.describe()  # says dataset statistics
    obj_df = x.select_dtypes(include=['object']).copy()  # takes everything that is an object (not float or int) and puts it in a new dataset
    obj_df.head()  # prints the new dataset
    int_df = x.select_dtypes(include=['int64']).copy()  # takes everything that is an int and puts it in a new dataset
    float_df = x.select_dtypes(include=['float64']).copy()  # takes everything that is a float and puts it in a new dataset
    df_int_float = pd.concat([float_df, int_df], axis=1)
    variables = []  # starts a list that will store all variables that are not numbers
    for i in range(1, len(condensed_data)):  # goes through each variable
        for j in range(len(condensed_data[0])):  # in the 2D array
            try:  # if the value of the field can be turned into a float (is numerical)
                float(condensed_data[i][j])  # this means it's a number
            except ValueError:  # if it can't be (is a string)
                if condensed_data[0][j] not in variables:  # adds the variable name to the list if it isn't there already
                    variables.append(condensed_data[0][j])
    le = preprocessing.LabelEncoder()  # anything involving le shows the encoding of categorical variables
    for i in range(len(variables)):
        le.fit(obj_df[variables[i]].astype(str))
    obj_df_trf = obj_df.astype(str).apply(le.fit_transform)  # transforms the categorical variables into numbers.
    global df_final #also used in linreg()
    if not obj_df_trf.empty:
        df_final = pd.concat([df_int_float, obj_df_trf], axis=1)  # join_axes=[df_int_float.index])
    else:
        df_final = df_int_float
    df_final.head()  # shows the final dataset with all the encoding
    print(df_final)  # prints the final dataset
    print()
    print("***********************************************************************************************************")
    print()
    if (o is not None):
        f = open(o,"a")
        f.write(df_final.to_string(header=True, index=True))
        f.write("\n\n***********************************************************************************************************")
        f.write("\n\nModel Results: ")
        f.close()

def linreg(): #actual linear regression
    print("Model Results: ")
    #printing the corrected model_string
    model_string = []
    model_string.append(dep_var)
    model_string.append(" ~ ")
    for i in range(0, len(full_model_variable_list)):
        model_string.append(full_model_variable_list[i])
        model_string.append(" + ")
    model_string.pop(-1)
    global full_model
    full_model = ''.join(model_string)
    print(full_model) #prints model
    print()
    print("***********************************************************************************************************")
    print()
    index = 0
    global levels #also used in contrasting()
    levels = []
    for i in range(len(condensed_data[0])):
        if c == condensed_data[0][i]:
            index = i
    for i in range(1,len(condensed_data)):
        if condensed_data[i][index] not in levels:
            levels.append(condensed_data[i][index])
    for i in range(len(levels)):
        levels[i] = i

    #Beginning of the linear regression
    global X
    global y
    if "*" in m:
        #correcting the format of the model string
        model_string = []
        model_string.append(dep_var)
        model_string.append(" ~ ")
        for i in range(0,len(full_model_variable_list)):
            model_string.append(full_model_variable_list[i])
            model_string.append(" + ")
        model_string.pop(-1)
        for i in range(0,len(model_string)):
            if "*" in model_string[i]:
                replacement = model_string[i].split("*")
                model_string[i] = replacement[0] + ":" + replacement[1]
               #makes sure the model is in the right format.
        string = ''.join(model_string)
        y, X = dmatrices(string, df_final)
    else:
        X = df_final[independentvariables]  # gets the modified values of the independent variables
        y = df_final[dep_var] # gets the modified values of the dependent variable
    if not c:
        #The linear regression
        regressor = LinearRegression()
        regressor.fit(X, y)
        regression = regressor.fit(X, y)
        #Data about the linear regression, starting without contrast
        X2 = sm.add_constant(X)
        statistics = sm.OLS(y, X2)
        finalstats = statistics.fit()
        print(finalstats.summary())
        if (o is not None):
            # concatenate data frames
            """f = open(o,"a")
            f.write(full_model)
            f.write("\n*************************************************************************************\n")
            f.write(finalstats.summary())
            f.close()"""
            sys.stdout = open(o, "a")
            print(full_model)
            print("\n*************************************************************************************\n")
            print(finalstats.summary())
            sys.stdout.close()
        return finalstats
def contrasting():
    global c
    if c:
        #to account for multiple contrast variables
        contrastvars = []
        if "," in c:
            contrastvars = c.split(",")
        for i in range(len(contrastvars)):
            contrastvars[i] = contrastvars[i].strip()
            if " " in contrastvars[i]:
                contrastvars[i]=contrastvars[i].replace(" ","_")
            if "/" in contrastvars[i]: #to account for URLs
                splitted = contrastvars[i].split("/")
                contrastvars[i] = splitted[len(splitted) - 1]
        else:
            splitted = c.split("/") #to account for URLs
            c = splitted[len(splitted) - 1]

        ind_vars_no_contrast_var = ''
        index = 1
        for i in range(len(full_model_variable_list)):
            if "/" in full_model_variable_list[i]:
                splitted = full_model_variable_list[i].split("/")
                full_model_variable_list[i] = splitted[len(splitted) - 1]
            if " " in full_model_variable_list[i]:
                full_model_variable_list[i]=full_model_variable_list[i].replace(" ","_")
        for var in full_model_variable_list:
            if var != c and not(var in contrastvars):
                if index == 1:
                    ind_vars_no_contrast_var = var
                    index += 1
                else:
                    ind_vars_no_contrast_var = ind_vars_no_contrast_var + " + " + var
        if len(contrastvars)>0:
            contraststring = ' + '.join(contrastvars)
        else:
            if " " in c:
                c = c.replace(" ", "_")
            contraststring=c
        # With contrast (treatment coding)
        print("\n\nTreatment (Dummy) Coding: Dummy coding compares each level of the categorical variable to a base reference level. The base reference level is the value of the intercept.")
        ctrst = Treatment(reference=0).code_without_intercept(levels)
        mod = ols(dep_var + " ~ " + ind_vars_no_contrast_var + " + C(" + contraststring + ", Treatment)", data=df_final)
        res = mod.fit()
        print("With contrast (treatment coding)")
        print(res.summary())
        if (o is not None):
            # concatenate data frames
            f = open(o, "a")
            f.write("\n" + full_model)
            f.write(
                "\n\n***********************************************************************************************************")

            f.write("\n\n\n\nTreatment (Dummy) Coding: Dummy coding compares each level of the categorical variable to a base reference level. The base reference level is the value of the intercept.")
            f.write("With contrast (treatment coding)")
            f.write(res.summary().as_text())
            f.close()
        # Defining the Simple class
        def _name_levels(prefix, levels):
            return ["[%s%s]" % (prefix, level) for level in levels]

        class Simple(object):
            def _simple_contrast(self, levels):
                nlevels = len(levels)
                contr = -1. / nlevels * np.ones((nlevels, nlevels - 1))
                contr[1:][np.diag_indices(nlevels - 1)] = (nlevels - 1.) / nlevels
                return contr

            def code_with_intercept(self, levels):
                c = np.column_stack((np.ones(len(levels)), self._simple_contrast(levels)))
                return ContrastMatrix(c, _name_levels("Simp.", levels))

            def code_without_intercept(self, levels):
                c = self._simple_contrast(levels)
                return ContrastMatrix(c, _name_levels("Simp.", levels[:-1]))

        ctrst = Simple().code_without_intercept(levels)
        mod = ols(dep_var + " ~ " + ind_vars_no_contrast_var + " + C(" + contraststring + ", Simple)", data=df_final)
        res = mod.fit()
        print("\n\nSimple Coding: Like Treatment Coding, Simple Coding compares each level to a fixed reference level. However, with simple coding, the intercept is the grand mean of all the levels of the factors.")
        print(res.summary())
        if (o is not None):
            # concatenate data frames
            f = open(o, "a")
            f.write("\n\n\nSimple Coding: Like Treatment Coding, Simple Coding compares each level to a fixed reference level. However, with simple coding, the intercept is the grand mean of all the levels of the factors.")
            f.write(res.summary().as_text())
            f.close()

        #With contrast (sum/deviation coding)
        ctrst = Sum().code_without_intercept(levels)
        mod = ols(dep_var + " ~ " + ind_vars_no_contrast_var + " + C(" + contraststring + ", Sum)", data=df_final)
        res = mod.fit()
        print("\n\nSum (Deviation) Coding: Sum coding compares the mean of the dependent variable for a given level to the overall mean of the dependent variable over all the levels.")
        print(res.summary())
        if (o is not None):
            # concatenate data frames
            f = open(o, "a")
            f.write("\n\n\nSum (Deviation) Coding: Sum coding compares the mean of the dependent variable for a given level to the overall mean of the dependent variable over all the levels.")
            f.write(res.summary().as_text())
            f.close()

        #With contrast (backward difference coding)
        ctrst = Diff().code_without_intercept(levels)
        mod = ols(dep_var + " ~ " + ind_vars_no_contrast_var + " + C(" + contraststring + ", Diff)", data=df_final)
        res = mod.fit()
        print("\n\nBackward Difference Coding: In backward difference coding, the mean of the dependent variable for a level is compared with the mean of the dependent variable for the prior level.")
        print(res.summary())
        if (o is not None):
            # concatenate data frames
            f = open(o, "a")
            f.write("\n\n\nBackward Difference Coding: In backward difference coding, the mean of the dependent variable for a level is compared with the mean of the dependent variable for the prior level.")
            f.write(res.summary().as_text())
            f.close()

        #With contrast (Helmert coding)
        ctrst = Helmert().code_without_intercept(levels)
        mod = ols(dep_var + " ~ " + ind_vars_no_contrast_var + " + C(" + contraststring + ", Helmert)", data=df_final)
        res = mod.fit()
        print("\n\nHelmert Coding: Our version of Helmert coding is sometimes referred to as Reverse Helmert Coding. The mean of the dependent variable for a level is compared to the mean of the dependent variable over all previous levels. Hence, the name ‘reverse’ being sometimes applied to differentiate from forward Helmert coding.")
        print(res.summary())
        if (o is not None):
            # concatenate data frames
            f = open(o, "a")
            f.write("\n\n\nHelmert Coding: Our version of Helmert coding is sometimes referred to as Reverse Helmert Coding. The mean of the dependent variable for a level is compared to the mean of the dependent variable over all previous levels. Hence, the name ‘reverse’ being sometimes applied to differentiate from forward Helmert coding.")
            f.write(res.summary().as_text())
            f.close()

def regularizing():
    if (r== ("L1" or "Lasso" or "l1" or "lasso") and not("y" in answer.lower())): #does it say L1, and has the user chosen to go ahead with running the code?
        # Loop to compute the cross-validation scores
        max_cross_val_alpha = 1
        max_cross_val_score = -1000000000.000 #making it a super negative number initially
        for x in range(1, MAX_ALPHA):
            lassoModel = Lasso(alpha=x, tol=0.0925)
            lassoModel.fit(X, y)
            scores = cross_val_score(lassoModel, X, y, cv=10)
            avg_cross_val_score = mean(scores) * 100
            #figure out which setting of the regularization parameter results in the max likelihood score
            if avg_cross_val_score > max_cross_val_score:
                max_cross_val_alpha = x
                max_cross_val_score = avg_cross_val_score

        # Building and fitting the Lasso Regression Model
        lassoModelChosen = Lasso(alpha=max_cross_val_alpha, tol=0.0925)
        lassoModelChosen.fit(X, y)
        print("\nLasso regression model:")
        print("Alpha with maximum likelihood (range: 1 to %d) = %f" %(MAX_ALPHA, max_cross_val_alpha))
        print("Current Model Score = %f" %(lassoModelChosen.score(X, y)))
        index = 0
        print("\nCoefficients:")
        if (o is not None):
            # concatenate data frames
            f = open(o, "a")
            f.write("\n\nLasso regression model:")
            f.write("\nAlpha with maximum likelihood (range: 1 to %d) = %f" %(MAX_ALPHA, max_cross_val_alpha))
            f.write("\nCurrent Model Score = %f" %(lassoModelChosen.score(X, y)))
            f.write("\n\nCoefficients:")
            f.close()
        for var in full_model_variable_list:
            print("%s \t %f" %(var,lassoModelChosen.coef_[index]))
            if (o is not None):
                with open(o, "a") as f:
                    f.write("\n%s \t %f" %(var, lassoModelChosen.coef_[index]))
                f.close()
            index = index + 1
        print("Intercept: %f" %(lassoModelChosen.intercept_))
        if (o is not None):
            with open(o, "a") as f:
                f.write("\nIntercept: %f" %(lassoModelChosen.intercept_))
                f.close()
        print()

    if (r== ("L2" or "Ridge" or "l2" or "Ridge") and not("y" in answer.lower())): #does it say L2, and has the user chosen to go ahead with running the code?
        # Loop to compute the different values of cross-validation scores
        max_cross_val_alpha = 1
        max_cross_val_score = -1000000000.000  # making it a super negative number initially
        for x in range(1, MAX_ALPHA):
            ridgeModel = Ridge(alpha=x, tol=0.0925)
            ridgeModel.fit(X, y)
            scores = cross_val_score(ridgeModel, X, y, cv=10)
            avg_cross_val_score = mean(scores) * 100
            # figure out which setting of the regularization parameter results in the max likelihood score
            if avg_cross_val_score > max_cross_val_score:
                max_cross_val_alpha = x
                max_cross_val_score = avg_cross_val_score

        # Building and fitting the Lasso Regression Model
        ridgeModelChosen = Ridge(alpha=max_cross_val_alpha, tol=0.0925)
        ridgeModelChosen.fit(X, y)
        print("\nRidge regression model:")
        print("Alpha with maximum likelihood (range: 1 to %d) = %f" % (MAX_ALPHA, max_cross_val_alpha))
        print("Current Model Score = %f" % (ridgeModelChosen.score(X, y)))
        index = 0
        """This numpy_conversion part was necessary because for the ridge model, all the coefficients get stored in a
        numpy array, and the conversion is necessary to get the coefficients. However, it is only needed if the model
        has interacting variables."""
        numpy_conversion = False
        for var in full_model_variable_list:
            if ("*" in var) or (":" in var):
                numpy_conversion = True
        if (o is not None):
            # concatenate data frames
            f = open(o, "a")
            f.write("\n\nRidge regression model:")
            f.write("\nAlpha with maximum likelihood (range: 1 to %d) = %f" %(MAX_ALPHA, max_cross_val_alpha))
            f.write("\nCurrent Model Score = %f" %(ridgeModelChosen.score(X, y)))
            f.write("\n\nCoefficients:")
            f.close()
        print("\nCoefficients:")
        if numpy_conversion:
            coeff_list = ridgeModelChosen.coef_[index].tolist()
            coeff_list.pop(0)
            for var in full_model_variable_list:
                print("%s \t %f" %(var, coeff_list[index]))
                if (o is not None):
                    with open(o,"a") as f:
                        f.write("\n%s \t %f" % (var, coeff_list[index]))
                index = index + 1
            print("Intercept: %f" % (ridgeModelChosen.intercept_))
            if (o is not None):
                with open(o, "a") as f:
                    f.write("\nIntercept: %f" % (ridgeModelChosen.intercept_))
                f.close()
            print()
        else:
            for var in full_model_variable_list:
                print("%s \t %f" % (var, ridgeModelChosen.coef_[index]))
                if (o is not None):
                    with open(o,"a") as f:
                        f.write("\n%s \t %f" % (var, ridgeModelChosen.coef_[index]))
                index = index + 1
            print("Intercept: %f" % (ridgeModelChosen.intercept_))
            if (o is not None):
                with open(o, "a") as f:
                    f.write("\nIntercept: %f" % (ridgeModelChosen.intercept_))
                f.close()
            print()
def opencsv(data):
    """saves a list of lists as a csv and opens"""
    import tempfile
    import os
    import csv
    handle, fn = tempfile.mkstemp(suffix='.csv')
    with os.fdopen(handle,"w", encoding='utf8',errors='surrogateescape',newline='') as f:
        writer = csv.writer(f)
        writer.writerows(data)
    return fn

# it can be used calling the script `python nidm_query.py -nl ... -q ..
if __name__ == "__main__":

    linear_regression()




import nidm.experiment.Navigate
from nidm.core import Constants
from tempfile import gettempdir
from nidm.experiment import Navigate
import nidm.experiment.tools.rest

import sys

from joblib import Memory
memory = Memory(gettempdir(), verbose=0)
USE_JOBLIB_CACHE = False


def GetProjectsComputedMetadata(nidm_file_list):
    '''
     :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :return: Dictionay or projects, each project having a dictionary of project stats
             including age_max, age_min, gender list, and handedness list.
    '''

    meta_data = {"projects": {}}
    projects = Navigate.getProjects(tuple(nidm_file_list))
    for p in projects:
        proj_id = nidm.experiment.tools.rest.RestParser.getTailOfURI(str(p))
        meta_data["projects"][proj_id] = {"age_max": 0, "age_min": sys.maxsize, "gender": [], "handedness": [] }
        meta_data["projects"][proj_id].update(Navigate.GetProjectAttributes(tuple(nidm_file_list), p))
        gender_set = set()
        hand_set = set()
        subjects = Navigate.getSubjects(tuple(nidm_file_list), p)
        for s in subjects:
            activities = Navigate.getActivities(tuple(nidm_file_list), s)
            meta_data["projects"][proj_id]["number_of_subjects"] = len(subjects)

            for a in activities:
                data = Navigate.getActivityData(tuple(nidm_file_list), a)
                if type(data) == nidm.experiment.Navigate.ActivityData:
                    for x in data.data:
                        if x.isAbout == Constants.NIDM_IS_ABOUT_AGE:
                            if float(x.value) > meta_data["projects"][proj_id]["age_max"]:
                                meta_data["projects"][proj_id]["age_max"] = float(x.value)
                            if float(x.value) < meta_data["projects"][proj_id]["age_min"]:
                                meta_data["projects"][proj_id]["age_min"] = float(x.value)
                        if x.isAbout == Constants.NIDM_IS_ABOUT_GENDER:
                            gender_set.add(str(x.value))
                        if x.isAbout == Constants.NIDM_IS_ABOUT_HANDEDNESS:
                            hand_set.add(str(x.value))
        meta_data["projects"][proj_id]["gender"] = list(gender_set)
        meta_data["projects"][proj_id]["handedness"] = list(hand_set)


    return meta_data
    # meta_data = GetProjectsMetadata(nidm_file_list)
    # ExtractProjectSummary(meta_data, nidm_file_list)

    # return compressForJSONResponse(meta_data)



import nidm.experiment.Navigate
from nidm.experiment import Query
import nidm.experiment.tools.rest_statistics
from nidm.core import Constants
import json
import re
from urllib import parse
import logging
import pprint
import os
from tempfile import gettempdir
from tabulate import tabulate
from copy import copy, deepcopy
from urllib.parse import urlparse, parse_qs
from  nidm.experiment import Navigate
from nidm.experiment.Utils import validate_uuid

from numpy import std, mean, median
import functools
import operator

from joblib import Memory
memory = Memory(gettempdir(), verbose=0)
USE_JOBLIB_CACHE = False

import simplejson

def convertListtoDict(lst):
    '''
    This function converts a list to a dictionary
    :param lst: list to convert
    :return: dictionary
    '''
    res_dct = {lst[i]: lst[i+1] for i in range(0,len(lst),2)}
    return res_dct
class RestParser:


    OBJECT_FORMAT = 0
    JSON_FORMAT = 1
    CLI_FORMAT = 2

    def __init__(self, verbosity_level = 0, output_format = 0):
        self.verbosity_level = verbosity_level
        self.output_format = output_format
        self.restLog ("Setting output format {}".format(self.output_format), 4)

    def setOutputFormat(self, output_format):
        self.output_format = output_format
        self.restLog ("Setting output format {}".format(self.output_format), 4)

    #####################
    # Standard formatters
    #####################

    def arrayFormat(self, result, headers):

        def allUUIDs(arr):
            uuid_only = True
            for s in arr:
                if type(s) != str or not re.match("^[0-9a-f]+-[0-9a-f]+-[0-9a-f]+-[0-9a-f]+-[0-9a-f]+$", s):
                    uuid_only = False
            return uuid_only


        if self.output_format == RestParser.JSON_FORMAT:
            return json.dumps(result, indent=2)
        elif self.output_format == RestParser.CLI_FORMAT:
            # most likely this is an array of strings but tabulate wants an array of arrays
            table = []
            for s in result:
                table.append( [s] )
            if allUUIDs(result) and headers[0] == "":
                headers[0] = "UUID"
            return tabulate(table, headers=headers)
        return result

    def dictFormat(self, result, headers=[""]):
        if self.output_format == self.CLI_FORMAT:
            table = []
            appendicies = []
            for key in result:

                # format a list
                if type(result[key]) == list:
                    appendix = []
                    for line in result[key]:
                        appendix.append( [ json.dumps(line) ] )
                    appendicies.append(tabulate(appendix, [key]))

                    # also put really short lists in as comma separated values
                    if len ( json.dumps(result[key]) ) < 40:
                        table.append( [ json.dumps(key), ",".join(result[key]) ] )

                # format a string
                elif type(result[key]) == str:
                    table.append([ json.dumps(key), result[key]])

                # format a dictionary
                elif type(result[key]) == dict:
                    # put any dict into it's own table at the end (sort of like an appendix)
                    appendix = []
                    for inner_key in result[key]:
                        appendix.append( [key, inner_key, result[key][inner_key] ] )
                    appendicies.append(tabulate(appendix))

                # format anything else
                else:
                    col1 = json.dumps(key)
                    if type(result[key]) == set:
                        col2 = json.dumps(list(result[key]))
                    else:
                        col2 = json.dumps(result[key])
                    table.append([col1, col2])

            return tabulate(table, headers) + "\n\n" + "\n\n".join(appendicies)
        else:
            return self.format(result)

    def objectTableFormat(self,result, headers = None):

        def flatten(obj, maxDepth=10, table = [], rowInProgress = [], depth = 0):
            for key in obj:
                newrow = deepcopy(rowInProgress)
                if depth< maxDepth and type(obj[key]) == dict:
                    newrow.append(key)
                    flatten(obj[key], maxDepth, table, newrow, depth+1)
                elif type(obj[key]) == str:
                    newrow.append(key)
                    newrow.append(obj[key])
                    table.append(newrow)
                else:
                    newrow.append(json.dumps(obj[key]))
                    table.append(newrow)

            return table

        if headers == None:
            headers = [""]

        return (tabulate(flatten(result), headers=headers))


    def activityDataTableFormat(self, data):
        headers = ['uuid', 'measure', 'label', 'value', 'unit']
        rows=[]
        for inst_or_deriv in data:
            for d in inst_or_deriv.data:
                rows.append( [inst_or_deriv.uuid, d.measureOf, d.label, d.value, d.hasUnit] )

        return (tabulate(rows, headers=headers))


    #####################
    # Custom formatters
    #####################

    def projectSummaryFormat(self, result):

        if self.output_format == self.CLI_FORMAT:
            ### added by DBK to sort things
            if "subjects" in result:
                result["subjects"]["uuid"],result["subjects"]["subject id"] = self.sort_list(result["subjects"]["uuid"], result["subjects"]["subject id"])
            else:
                result["subjects"] = []
            if "data_elements" in result:
                result["data_elements"]["uuid"],result["data_elements"]["label"] = self.sort_list(result["data_elements"]["uuid"], result["data_elements"]["label"])
            else:
                result["data_elements"] = []

            toptable = []
            for key in result:
                if not key in ['subjects', 'data_elements', 'field_values']:
                    toptable.append([ key, simplejson.dumps(result[key]) ])

            if 'field_values' in result and len(result['field_values']) > 0 :
                fh_header = ['subject', 'label', 'value', 'unit', 'isAbout'] #result['field_values'][0].keys()
                fh_rows = [ [x.subject, x.label, x.value, x.hasUnit, x.isAbout] for x in result['field_values']]
                field_table = tabulate(fh_rows, fh_header)
                #added by DBK, if they asked for fields then just give them the fields
                return "{}".format(field_table)
            else:
                field_table = ''

            return "{}\n\n{}\n{}\n\n{}\n{}\n\n{}".format(
                tabulate(toptable),
                ### modified by DBK to account for new dictionary format of results
                # tabulate({"subjects": result["subjects"]}, headers="keys"),
                # sort list 2 by list 1 and replace unsorted version
                tabulate([],headers=["Subject Information"]),
                tabulate(result["subjects"], headers="keys"),
                #tabulate({"data_elements": result["data_elements"]}, headers="keys"),
                tabulate([],headers = ["Data Elements"]),
                tabulate({'uuid': result["data_elements"]['uuid'], 'label': result["data_elements"]['label']}, headers="keys"),
                field_table
            )
        else:
            # added by DBK to check if we had fields requested then we should just return those
            if 'field_values' in result:
                # convert result['field_values'] to a list for json export
                return self.format(result['field_values'])
            else:
                return self.format(result)

    def formatDerivatives(self, derivative):
        self.restLog("formatting derivatives in format {}".format(self.output_format), 5)
        if self.output_format == self.CLI_FORMAT:
            table = []
            for uri in derivative:
                for measurement in derivative[uri]["values"]:
                    if measurement not in ["http://www.w3.org/ns/prov#wasGeneratedBy", "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"]:  # skip some NIDM structure artifacts
                        table.append([uri,
                                      measurement,
                                      derivative[uri]["values"][measurement]["label"],
                                      "{} {}".format(derivative[uri]["values"][measurement]["value"], derivative[uri]["values"][measurement]["units"]),
                                      derivative[uri]["values"][measurement]["datumType"],
                                      derivative[uri]["values"][measurement]["isAbout"]])
            return tabulate(table, headers=["Derivative_UUID", "Measurement", "Label", "Value", "Datumtype", "isAbout"])
        else:
            return self.format(derivative)

    def dataElementsFormat(self, de_data):
        if self.output_format == self.CLI_FORMAT:

            table = []
            headers = ['label', 'source_variable', 'hasUnit', 'description', 'dataElement', 'isAbout']

            # for each data element, create a row with each value from the header
            for de in de_data['data_elements']['data_type_info']:
                row = []
                for h in headers:
                    row.append(de[h])
                table.append(row)

            # print (de_data['data_elements']['data_type_info'][0])

            # Sort by the first column, which is label
            sorted_table = sorted(table, key=lambda x: x[0])

            text = tabulate(sorted_table, headers)
            return text
        else:
            return self.format(de_data)

    def dataElementDetailsFormat(self, de_data):
            return self.format(de_data)


    def subjectFormat(self, subject_data):
        if self.output_format == self.CLI_FORMAT:

            subjects = []
            for subject in subject_data['subject']:
                subjects.append(subject)
            text = tabulate(subjects, headers=["Subject UUID", "Source Subject ID"])

            if 'fields' in subject_data:
                field_data = []
                text += "\n\n"
                for sub in subject_data['fields']:
                    for act in subject_data['fields'][sub]:
                        de = subject_data['fields'][sub][act]
                        field_data.append([sub, act, de.label, de.value])
                text += tabulate(field_data, headers=["Subject", "Activity", "Field", "Value"])

            return text
        else:
            return self.format(subject_data)

    def subjectSummaryFormat(self,result):
        if self.output_format == self.CLI_FORMAT:
            special_keys = ['instruments', 'derivatives', "activity"]
            toptable = []
            for key in result:
                if not key in special_keys:
                    toptable.append([ key, result[key] ])

            for key in special_keys:
                if type(result[key]) == dict:
                    toptable.append( [ key, ",".join(result[key].keys()) ] )
                elif type(result[key]) == list:
                    toptable.append( [ key, ",".join(result[key]) ])
                else:
                    toptable.append([key, json.dumps(result[key]) ])

            instruments = self.objectTableFormat(result['instruments'], ["Instrument_UUID", "Category", "Value"])
            derivatives=  self.formatDerivatives(result['derivatives'])

            return "{}\n\n{}\n\n{}".format(
                tabulate(toptable),
                derivatives,
                instruments
            )
        else:
            return self.format(result)


    def subjectSummaryFormat_v2(self,result):
        if self.output_format == self.CLI_FORMAT:
            special_keys = ['instruments', 'derivatives', "activity"]
            toptable = []
            for key in result:
                if not key in special_keys:
                    toptable.append([ key, result[key] ])

            for key in special_keys:
                if key in result:
                    if type(result[key]) == dict:
                        toptable.append( [ key, ",".join(result[key].keys()) ] )
                    if type(result[key]) == list and len(result[key]) > 0 and  type(result[key][0]) == Navigate.ActivityData:
                        toptable.append( [ key, ",".join( [x.uuid for x in result[key] ]   ) ])
                    elif type(result[key]) == list:
                        toptable.append([key, ",".join])
                    else:
                        toptable.append([key, json.dumps(result[key]) ])

            instruments = self.activityDataTableFormat(result['instruments'])
            derivatives=  self.activityDataTableFormat(result['derivatives'])

            return "{}\n\n{}\n\n{}".format(tabulate(toptable), derivatives, instruments )
        else:
            return self.format(result)

    ### Added by DBK to sorty data elements lists
    #####################
    # Sort Functions
    #####################
    def sort_list (self,list1,list2):
        '''
        This function will sort list 1 using list 2 values, returning sorted list 1, sorted list 2
        '''

        if len(list1) == 0 or len(list2) == 0:
            return list1, list2

        list1 = list(zip(*sorted(zip(list2,list1))))[1]
        return list1,sorted(list2)

    #####################
    # Route Functions
    #####################

    def dataelements(self):
        result = Navigate.GetDataelements(self.nidm_files)
        return self.dataElementsFormat(result)

    def dataelementsSummary(self):
        path = (urlparse(self.command)).path
        match = re.match(r"^/?dataelements/([^/\?]+)", path)
        dataelement = parse.unquote(str(match.group(1)))

        result = Navigate.GetDataelementDetails(self.nidm_files, dataelement)
        return self.dataElementDetailsFormat(result)


    def projects(self):
        result = []
        field_values = []
        self.restLog("Returning all projects", 2)
        projects = Query.GetProjectsUUID(self.nidm_files)
        for uuid in projects:
            result.append(str(uuid).replace(Constants.NIIRI, ""))


        # if we got fields, drill into each subject and pull out the field data
        # subject details -> derivitives / instrument -> values -> element
        if 'fields' in self.query and len(self.query['fields']) > 0:
            subjects_set = set()
            dataelements_set = set()
            self.restLog("Using fields {}".format(self.query['fields']), 2)
            # result['field_values'] = []

            for proj in projects:
                # get all the synonyms for all the fields
                field_synonyms = functools.reduce(operator.iconcat,
                                                  [Query.GetDatatypeSynonyms(self.nidm_files, proj, x) for x in
                                                   self.query['fields']], [])

                files = self.nidm_files
                all_subjects = Query.GetParticipantUUIDsForProject(self.nidm_files, proj, self.query['filter']) # nidm_file_list= files, project_id=proj['uuid'], filter=self.query['filter']):
                for sub in all_subjects['uuid']:

                    for activity in Navigate.getActivities(self.nidm_files, sub):
                        activity = Navigate.getActivityData(self.nidm_files, activity)
                        for data_element in activity.data:
                            if data_element.dataElement in field_synonyms:
                                field_values.append(data_element._replace(subject=sub))
                                subjects_set.add(sub)
                                dataelements_set.add( (data_element.datumType, data_element.label) )

            if len(field_values) == 0:
                raise ValueError("Supplied field not found. (" + ", ".join(self.query['fields']) + ")")
            else:
                summary_result = {}
                summary_result['subjects']= {"uuid":[], "subject id":[]}
                for sub in subjects_set:
                    summary_result["subjects"]["uuid"].append(sub)
                    summary_result["subjects"]["subject id"].append("")
                summary_result['data_elements'] = {"uuid": [], "lable":[]}
                for de in dataelements_set:
                    summary_result['data_elements']["uuid"] = de[0]
                    summary_result['data_elements']["label"] = de[1]
                summary_result['field_values'] = field_values
                return self.projectSummaryFormat(summary_result)


        return self.format(result, ["UUID"])

    def ExpandProjectMetaData(self, meta_data):
        """
        Takes in the meta_data from GetProjectsMetadata() and adds
        the following statistics about each project to the existing
        meta_data structure:
         age_max (float)
         age_min (float)
         handedness of subjects (list)
         genders of subjects (list)
         number of subjects (int)

        :param meta_data:
        :return:
        """
        for project_id in meta_data['projects']:

            project_uuid = str(project_id)[6:] if (str(project_id).startswith("niiri:")) else project_id
            project = meta_data['projects'][project_id]

            ages = set()
            hands = set()
            genders = set()


            for session in Navigate.getSessions(self.nidm_files, project_uuid):
                for acq in Navigate.getAcquisitions(self.nidm_files, session):
                    act_data = Navigate.getActivityData(self.nidm_files, acq)
                    for de in act_data.data:
                        if de.isAbout == "http://uri.interlex.org/ilx_0100400" or de.isAbout == "http://uri.interlex.org/base/ilx_0100400":
                            ages.add(float(de.value))
                        elif de.isAbout == "http://uri.interlex.org/ilx_0101292" or de.isAbout == "http://uri.interlex.org/base/ilx_0101292"\
                                or de.isAbout == "http://uri.interlex.org/ilx_0738439" or de.isAbout == \
                                "https://ndar.nih.gov/api/datadictionary/v2/dataelement/gender":
                            genders.add(de.value)
                        elif de.isAbout == "http://purl.obolibrary.org/obo/PATO_0002201":
                            hands.add(de.value)

            print(Query.GetParticipantUUIDsForProject(self.nidm_files, project_uuid))

            project['age_max'] = max(ages) if len(ages) > 0 else 0
            project['age_min'] = min(ages) if len(ages) > 0 else 0
            project[Query.matchPrefix(str(Constants.NIDM_NUMBER_OF_SUBJECTS))] = len((Query.GetParticipantUUIDsForProject(self.nidm_files, project_uuid))['uuid'])
            project[str(Constants.NIDM_GENDER)] = list(genders)
            project[str(Constants.NIDM_HANDEDNESS)] = list(hands)

    def projectStats(self):
        result = dict()
        subjects = None
        path = (urlparse(self.command)).path

        match = re.match(r"^/?statistics/projects/([^/]+)\??$", path)
        id = parse.unquote(str(match.group(1)))
        self.restLog("Returing project {} stats metadata".format(id), 2)

        meta_data = Query.GetProjectsMetadata(self.nidm_files)
        self.ExpandProjectMetaData(meta_data)
        projects = Query.compressForJSONResponse(meta_data)

        for pid in projects['projects'].keys():
            self.restLog("comparng " + str(pid) + " with " + str(id), 5)
            self.restLog("comparng " + str(pid) + " with " + Constants.NIIRI + id, 5)
            self.restLog("comparng " + str(pid) + " with niiri:" + id, 5)
            if pid == id or pid == Constants.NIIRI + id or pid == "niiri:" + id:
                # stip off prefixes to make it more human readable
                for key in projects['projects'][pid]:
                    short_key = key
                    possible_prefix = re.sub(':.*', '', short_key)
                    if possible_prefix in Constants.namespaces:
                        short_key = re.sub('^.*:', '', short_key)
                    result[short_key] = projects['projects'][pid][key]

        # now get any fields they reqested
        for field in self.query['fields']:
            if subjects == None:
                subjects = Query.GetParticipantUUIDsForProject(tuple(self.nidm_files), project_id=id, filter=self.query['filter'])
                result['subjects'] = subjects['uuid']
            bits = field.split('.')
            if len(bits) > 1:
                stat_type = self.getStatType(bits[0]) # should be either instruments or derivatives for now.
                self.addFieldStats(result, id, subjects['uuid'], bits[1], stat_type) # bits[1] will be the ID

        return self.dictFormat(result)

    STAT_TYPE_OTHER = 0
    STAT_TYPE_INSTRUMENTS = 1
    STAT_TYPE_DERIVATIVES = 2
    def getStatType(self, name):
        lookup = {"instruments": self.STAT_TYPE_INSTRUMENTS, "derivatives" : self.STAT_TYPE_DERIVATIVES}
        if name in lookup: return lookup[name]
        return self.STAT_TYPE_OTHER


    @staticmethod
    def getTailOfURI(uri):
        if '#' in uri:
            return uri[uri.rfind('#') + 1:]
        else:
            return uri[uri.rfind('/') + 1:]


    def addFieldStats(self, result, project, subjects, field, type):
        '''
        Geneerates basic stats on a group of subjects and adds it to the result
        :param result:
        :param subjects:
        :param field:
        :return:
        '''
        values = []
        for s in subjects:
            if type == self.STAT_TYPE_INSTRUMENTS:
                data = Query.GetParticipantInstrumentData(tuple(self.nidm_files), project, s)
                for i in data:
                    if field in data[i]:
                        values.append( float(data[i][field]) )
            # derivatives are of the form [UUID]['values'][URI]{datumType, label, values, units}
            if type == self.STAT_TYPE_DERIVATIVES:
                data = Query.GetDerivativesDataForSubject(tuple(self.nidm_files), project, s)
                for deriv in data:
                    for URI in data[deriv]['values']:
                        measures = data[deriv]['values'][URI]
                        if field == measures['label'] or field == self.getTailOfURI(URI):
                            values.append( float(measures['value']) )

        if len(values) > 0:
            med = median(values)
            avg = mean(values)
            st = std(values)
            mn = min(values)
            mx = max(values)
        else:
            med = avg = st = mn = mx = None
        result[field] = {"max": mx, "min": mn, "median": med, "mean": avg, "standard_deviation": st}

    def projectSummary(self):

        match = re.match(r"^/?projects/([^/]+)$", self.command)
        id = parse.unquote(str(match.group(1)))
        self.restLog("Returing project {} summary".format(id), 2)

        result = nidm.experiment.Navigate.GetProjectAttributes(self.nidm_files, project_id=id)
        result['subjects'] = Query.GetParticipantUUIDsForProject(self.nidm_files, project_id=id, filter=self.query['filter'])
        result['data_elements'] = Query.GetProjectDataElements(self.nidm_files, project_id=id)


        # if we got fields, drill into each subject and pull out the field data
        # subject details -> derivitives / instrument -> values -> element
        if 'fields' in self.query and len(self.query['fields']) > 0:
            self.restLog("Using fields {}".format(self.query['fields']), 2)
            result['field_values'] = []
            # get all the synonyms for all the fields
            field_synonyms = functools.reduce( operator.iconcat, [ Query.GetDatatypeSynonyms(self.nidm_files, id, x) for x in self.query['fields'] ], [])
            for sub in result['subjects']['uuid']:

                for activity in Navigate.getActivities(self.nidm_files, sub):
                    activity = Navigate.getActivityData(self.nidm_files, activity)
                    for data_element in activity.data:
                        if data_element.dataElement in field_synonyms:
                            result['field_values'].append(data_element._replace(subject=sub))

            if len(result['field_values']) == 0:
                raise ValueError("Supplied field not found. (" + ", ".join(self.query['fields']) + ")")

        return self.projectSummaryFormat(result)


    def subjectsList(self):
        match = re.match(r"^/?projects/([^/]+)/subjects/?$", self.command)
        project = match.group((1))
        self.restLog("Returning all agents matching filter '{}' for project {}".format(self.query['filter'], project), 2)
        # result = Query.GetParticipantUUIDsForProject(self.nidm_files, project, self.query['filter'], None)
        all_subjects = Navigate.getSubjects(self.nidm_files, project)
        result = {}
        result['uuid'] = []
        result['subject id'] = []
        for sub_uuid in all_subjects:
            if Query.CheckSubjectMatchesFilter(self.nidm_files,project, sub_uuid, self.query['filter']):
                uuid_string = (str(sub_uuid)).split('/')[-1]  # srip off the http://whatever/whatever/
                result['uuid'].append(uuid_string)
                sid = Navigate.getSubjectIDfromUUID(self.nidm_files, sub_uuid)
                result['subject id'].append(str(sid))
        return self.format(result)

    def projectSubjectSummary(self):
        match = re.match(r"^/?projects/([^/]+)/subjects/([^/]+)/?$", self.command)
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        self.restLog("Returning info about subject {}".format(match.group(2)), 2)
        return self.subjectSummaryFormat(Query.GetParticipantDetails(self.nidm_files, match.group(1), subject))

    def getFieldInfoForSubject(self, project, subject):
        '''
        Returns a dictionary of activities where the subject has matching field data
        The result[activity] is the full data_element so to get the value you would use result[activity].value
        Note that a subject could match the same field in multiple activities.

        :param project:
        :param subject:
        :return:
        '''
        result = {}
        # if we got fields, drill into each subject and pull out the field data
        # subject details -> derivitives / instrument -> values -> element
        if 'fields' in self.query and len(self.query['fields']) > 0:
            # get all the synonyms for all the fields - we can seach for them all at once
            field_synonyms = functools.reduce( operator.iconcat, [ Query.GetDatatypeSynonyms(self.nidm_files, project, x) for x in self.query['fields'] ], [])

            # print (field_synonyms)

            for activity in Navigate.getActivities(self.nidm_files, subject):
                activity_data = Navigate.getActivityData(self.nidm_files, activity)
                # print ([ x.label for x in activity.data])
                for data_element in activity_data.data:
                    if not set([data_element.dataElement, data_element.label, data_element.isAbout]).isdisjoint(set(field_synonyms)):
                        result[Query.URITail(activity)] = data_element
        return result


    def subjects(self):
        self.restLog("Returning info about subjects",2)
        projects = Navigate.getProjects(self.nidm_files)
        result = {'subject': []}
        if 'fields' in self.query and len(self.query['fields']) > 0:
            result['fields'] = {}

        for proj in projects:
            subs = Navigate.getSubjects(self.nidm_files, proj)
            for s in subs:
                result['subject'].append( [Query.URITail(s), Navigate.getSubjectIDfromUUID(self.nidm_files, s) ])

                # print ("getting info for " + str(s))
                x = self.getFieldInfoForSubject(proj, s)
                if x != {}:
                    result['fields'][Query.URITail(s)] = x
        return self.subjectFormat(result)

    def subjectSummary(self):
        match = re.match(r"^/?subjects/([^/]+)/?$", self.command)
        self.restLog("Returning info about subject {}".format(match.group(1)), 2)
        id = match.group(1)

        # if we were passed in a sub_id rather than a UUID, lookup the associated UUID. (we might get multiple!)
        if validate_uuid(id):
            sub_ids = id
        else:
            sub_ids = Navigate.getSubjectUUIDsfromID(self.nidm_files, id)
            if len(sub_ids) == 1:
                sub_ids = sub_ids[0]

        activities = Navigate.getActivities(self.nidm_files, id)
        activityData = []
        for a in activities:
            data = Navigate.getActivityData(self.nidm_files, a)
            activityData.append(data)

        return self.subjectSummaryFormat_v2( {'uuid': sub_ids,
                'instruments' : list(filter(lambda x: x.category == 'instrument', activityData)),
                'derivatives' : list(filter(lambda x: x.category == 'derivative', activityData))
                })


    def instrumentsList(self):
        result = []
        match = re.match(r"^/?projects/([^/]+)/subjects/([^/]+)/instruments/?$", self.command)
        self.restLog("Returning instruments in subject {}".format(match.group(2)), 2)
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        instruments = Query.GetParticipantInstrumentData(self.nidm_files, match.group(1), subject)
        for i in instruments:
            result.append(i)
        return self.format(result)

    def instrumentSummary(self):
        match = re.match(r"^/?projects/([^/]+)/subjects/([^/]+)/instruments/([^/]+)$", self.command)
        self.restLog("Returning instrument {} in subject {}".format(match.group(3), match.group(2)), 2)
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        instruments = Query.GetParticipantInstrumentData(self.nidm_files, match.group(1), subject)
        return self.format(instruments[match.group(3)], headers=["Category", "Value"])

    def derivativesList(self):
        result = []
        match = re.match(r"^/?projects/([^/]+)/subjects/([^/]+)", self.command)
        self.restLog("Returning derivatives in subject {}".format(match.group(2)), 2)
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        derivatives = Query.GetDerivativesDataForSubject(self.nidm_files, match.group(1), subject)
        for s in derivatives:
            result.append(s)
        return self.format(result)

    def derivativeSummary(self):
        match = re.match(r"^/?projects/([^/]+)/subjects/([^/]+)/derivatives/([^/]+)", self.command)
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        uri = match.group(3)
        self.restLog("Returning stat {} in subject {}".format(uri, match.group(2)), 2)
        derivatives = Query.GetDerivativesDataForSubject(self.nidm_files, match.group(1), subject)

        single_derivative = { uri: derivatives[uri] }

        self.restLog("Formatting single derivative", 5)


        return self.formatDerivatives(single_derivative)

    def run(self, nidm_files, command):
        try:
            self.restLog("parsing command " + command, 1)
            self.restLog("Files to read:" + str(nidm_files), 1)
            self.restLog("Using {} as the graph cache directory".format(gettempdir()), 1)

            self.nidm_files = tuple(nidm_files)
            #replace # marks with %23 - they are sometimes used in the is_about terms
            escaped = command.replace("#", "%23")
            u = urlparse(escaped)
            self.command = u.path
            self.query = parse_qs(u.query)

            if 'filter' in self.query:
                self.query['filter'] = self.query['filter'][0]
            else:
                self.query['filter'] = None

            # normalize query dict for our particular situation
            if 'fields' in self.query:
                self.query['fields'] = str.split(self.query['fields'][0], ',')
            else:
                self.query['fields'] = []

            return self.route()
        except ValueError as ve:
            logging.error("Exception: {}".format(ve))
            return (self.format({"error": "One of the supplied field terms was not found."}))



    def route(self):

        if re.match(r"^/?dataelements/?$", self.command): return self.dataelements()

        if re.match(r"^/?dataelements/[^/]+/?$", self.command): return self.dataelementsSummary()

        if re.match(r"^/?projects/?$", self.command): return self.projects()

        if re.match(r"^/?statistics/projects/[^/]+$", self.command): return self.projectStats()

        if re.match(r"^/?projects/[^/]+$", self.command): return self.projectSummary()

        if re.match(r"^/?subjects/?$", self.command): return self.subjects()

        if re.match(r"^/?subjects/[^/]+$", self.command): return self.subjectSummary()

        if re.match(r"^/?projects/[^/]+/subjects/?$", self.command): return self.subjectsList()

        if re.match(r"^/?projects/[^/]+/subjects/[^/]+/?$", self.command): return self.projectSubjectSummary()

        if re.match(r"^/?projects/[^/]+/subjects/[^/]+/instruments/?$", self.command): return self.instrumentsList()

        if re.match(r"^/?projects/[^/]+/subjects/[^/]+/instruments/[^/]+/?$", self.command): return self.instrumentSummary()

        if re.match(r"^/?projects/[^/]+/subjects/[^/]+/derivatives/?$", self.command): return self.derivativesList()

        if re.match(r"^/?projects/[^/]+/subjects/[^/]+/derivatives/[^/]+/?$", self.command): return self.derivativeSummary()

        self.restLog("NO MATCH!", 2)

        return {"error": "No match for supplied URI"}


    def restLog(self, message, verbosity_of_message):
        if verbosity_of_message <= self.verbosity_level:
            print (message)



    def format(self, result, headers = [""]):
        if self.output_format == RestParser.JSON_FORMAT:
            json_str = simplejson.dumps(result, indent=2)
            return json_str

        elif self.output_format == RestParser.CLI_FORMAT:
            if type(result) == dict:
                return self.dictFormat(result, headers)
            if type(result) == list:
                return self.arrayFormat(result, headers)
            else:
                return str(result)

        return result



from nidm.core import Constants
from nidm.experiment.Query import OpenGraph, URITail, trimWellKnownURIPrefix, getDataTypeInfo, ACQUISITION_MODALITY, \
    IMAGE_CONTRAST_TYPE, IMAGE_USAGE_TYPE, TASK, expandUUID, matchPrefix
from rdflib import Graph, RDF, URIRef, util, term, Literal
import functools
import collections
import nidm.experiment.CDE
from nidm.experiment.Utils import validate_uuid


isa = URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
isPartOf = Constants.DCT['isPartOf']
ValueType = collections.namedtuple('ValueType',
                                   ['value', 'label', 'datumType', 'hasUnit', 'isAbout', 'measureOf', 'hasLaterality', 'dataElement', 'description', 'subject', 'project', 'sourceVariable'])
ActivityData = collections.namedtuple('ActivityData', ['category', 'uuid', 'data'])
QUERY_CACHE_SIZE=0
BIG_CACHE_SIZE=0

def makeValueType(value=None, label=None, datumType=None, hasUnit=None, isAbout=None, measureOf=None, hasLaterality=None, dataElement=None, description=None, subject=None, project=None, source_variable=None):
    return ValueType(str(value), str(label), str(datumType), str(hasUnit), str(isAbout), str(measureOf), str(hasLaterality), str(dataElement), str(description), str(subject), str(project), str(source_variable))

def makeValueTypeFromDataTypeInfo(value, data_type_info_tuple):

    if not data_type_info_tuple:
        data_type_info_tuple = {}

    for key in ['label', 'datumType', 'hasUnit', 'isAbout', 'measureOf', 'hasLaterality', 'dataElement', 'description', 'subject', 'project', 'source_variable']:
        if not key in data_type_info_tuple:
            data_type_info_tuple[key] = None


    return ValueType(str(value), str(data_type_info_tuple['label']), str(data_type_info_tuple['datumType']),
                     str(data_type_info_tuple['hasUnit']), str(data_type_info_tuple['isAbout']), str(data_type_info_tuple['measureOf']),
                     str(data_type_info_tuple['hasLaterality']), str(data_type_info_tuple['dataElement']),
                     str(data_type_info_tuple['description']), str(data_type_info_tuple['subject']), str(data_type_info_tuple['project']), str(data_type_info_tuple['source_variable']))

def expandID(id, namespace):
    '''
    If the ID isn't a full URI already, make it one in the given namespace

    :param id:
    :param namespace:
    :return: full URI
    '''
    if id.find('http') < 0:
        return namespace[id]
    # it has a http, but isn't a URIRef so convert it
    if type(id) == str:
        return URIRef(id)

    return id


@functools.lru_cache(maxsize=BIG_CACHE_SIZE)
def simplifyURIWithPrefix(nidm_file_tuples, uri):
    '''
    Takes a URI and finds if there is a simple prefix for it in the graph
    :param rdf_graph:
    :param uri:
    :return: simple prefix or the original uri string
    '''

    @functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
    def getNamespaceLookup(nidm_file_tuples):
        names = {}
        for f in nidm_file_tuples:
            rdf_graph = OpenGraph(f)
            for (prefix, uri) in rdf_graph.namespace_manager.namespaces():
                if not str(uri) in names:
                    names[str(uri)] = prefix
        return names

    names = getNamespaceLookup(tuple(nidm_file_tuples))
    # strip off the bit of URI after the last /
    trimed_uri = str(uri).split('/')[0:-1]
    trimed_uri = '/'.join(trimed_uri) + '/'
    if trimed_uri in names:
        return names[trimed_uri]
    else:
        return uri

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getProjects(nidm_file_tuples):
    projects = []

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        #find all the sessions
        for (project, p, o) in rdf_graph.triples((None, isa, Constants.NIDM['Project'])):
            projects.append(project)

    return projects

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSessions(nidm_file_tuples, project_id):
    project_uri = expandID(project_id, Constants.NIIRI)
    sessions = []

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        #find all the sessions
        for (session, p, o) in rdf_graph.triples((None, isa, Constants.NIDM['Session'])):
            #check if it is part of our project
            if (session, isPartOf, project_uri) in rdf_graph:
                sessions.append(session)

    return sessions

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getAcquisitions(nidm_file_tuples, session_id):
    session_uri = expandID(session_id, Constants.NIIRI)
    acquisitions = []

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        #find all the sessions
        for (acq, p, o) in rdf_graph.triples((None, isPartOf, session_uri)):
            #check if it is a acquisition
            if (acq, isa, Constants.NIDM['Acquisition']) in rdf_graph:
                acquisitions.append(acq)

    return acquisitions

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSubject(nidm_file_tuples, acquisition_id):
    acquisition_uri = expandID(acquisition_id, Constants.NIIRI)
    subjects = []

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        #find all the sessions
        for (acq, p, blank) in rdf_graph.triples((acquisition_uri, Constants.PROV['qualifiedAssociation'], None)):
            for (s2, p2, sub) in rdf_graph.triples((blank, Constants.PROV['agent'], None)):
                if (blank, Constants.PROV['hadRole'], Constants.SIO['Subject']) in rdf_graph:
                    return sub
    return None

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSubjects(nidm_file_tuples, project_id):
    subjects = set([])
    project_uri = expandID(project_id, Constants.NIIRI)

    sessions = getSessions(nidm_file_tuples, project_uri)
    for s in sessions:
        acquisitions = getAcquisitions(nidm_file_tuples, s)
        for acq in acquisitions:
            sub = getSubject(nidm_file_tuples, acq)
            subjects.add(sub)
    return subjects

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSubjectUUIDsfromID(nidm_file_tuples, sub_id):
    uuids = []
    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)

        result = rdf_graph.triples((None, Constants.NDAR['src_subject_id'], None))
        for (s,p,o) in result:
            if str(o) == str(sub_id):
                uuids.append( URITail(s) )

    return uuids

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSubjectIDfromUUID(nidm_file_tuples, subject_uuid):
    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        id_generator = rdf_graph.objects(subject=subject_uuid, predicate=Constants.NDAR['src_subject_id'])
        for id in id_generator:
            return id
    return None

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def normalizeSingleSubjectToUUID(nidm_file_tuples, id):
    if len(getSubjectUUIDsfromID(nidm_file_tuples, id)) > 0:
        return getSubjectUUIDsfromID(nidm_file_tuples, id)[0]
    return id

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getActivities(nidm_file_tuples, subject_id):
    activities = set([])

    # if we were passed in a sub_id rather than a UUID, lookup the associated UUID. (we might get multiple!)
    if validate_uuid(URITail(subject_id)):
        sub_uris = [subject_id]
    else:
        sub_uris = getSubjectUUIDsfromID(nidm_file_tuples, subject_id)

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        for subject_uri in sub_uris:
            subject_uri = expandID(subject_uri, Constants.NIIRI)
            for blank_node in rdf_graph.subjects(predicate=Constants.PROV['agent'], object=subject_uri):
                for activity in rdf_graph.subjects(predicate=Constants.PROV['qualifiedAssociation'], object=blank_node):
                    if (activity, isa, Constants.PROV['Activity']) in rdf_graph:
                        activities.add(activity)
    return activities

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def isAStatCollection(nidm_file_tuples, uri):
    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        if ((uri, isa, Constants.NIDM['FSStatsCollection']) in rdf_graph ) or \
            ((uri, isa, Constants.NIDM['FSLStatsCollection']) in rdf_graph) or \
            ((uri, isa, Constants.NIDM['ANTSStatsCollection']) in rdf_graph) :
            return True
    return False

# def getDataElementInfo(nidm_file_list, id):
#
#     uuid = expandID(id, Constants.NIIRI)
#
#     for file in nidm_file_list:
#         rdf_graph = OpenGraph(file)
#         if (uuid, isa, Constants.NIDM['DataElement']) in rdf_graph:
#             label = list(rdf_graph.objects(subject=uuid, predicate=Constants.RDFS['label'])) [0]
#             description = list(rdf_graph.objects(subject=uuid, predicate=Constants.DCT['description'])) [0]
#             isAbout = list(rdf_graph.objects(subject=uuid, predicate=Constants.NIDM['isAbout'])) [0]
#             return makeValueType(label=label, description=description, isAbout=isAbout)
#
#     return False

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getActivityData(nidm_file_tuples, acquisition_id):
    acquisition_uri = expandID(acquisition_id, Constants.NIIRI)
    result = []
    category = None

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        # find everything generated by the acquisition
        for (data_object, p1, o1) in rdf_graph.triples((None, Constants.PROV['wasGeneratedBy'], acquisition_uri)):
            # make sure this is an acquisition object
            if (data_object, isa, Constants.NIDM['AcquisitionObject']) in rdf_graph:
                category = 'instrument'
                # iterate over all the items in the acquisition object
                for (s, p, o) in rdf_graph.triples((data_object, None, None)):

                    dti = getDataTypeInfo(rdf_graph, p)
                    if (dti):
                        # there is a DataElement describing this predicate
                        value_type = makeValueTypeFromDataTypeInfo(value=trimWellKnownURIPrefix(o), data_type_info_tuple=dti)
                        result.append( value_type )
                    else:
                        #Don't know exactly what this is so just set a label and be done.
                        if (data_object, isa, Constants.ONLI['assessment-instrument']) in rdf_graph:
                            result.append(makeValueType(value=trimWellKnownURIPrefix(o), label=simplifyURIWithPrefix(nidm_file_tuples, str(p))))
                            #result[ simplifyURIWithPrefix(nidm_file_list, str(p)) ] = trimWellKnownURIPrefix(o)
                        else:
                            result.append(makeValueType(value=trimWellKnownURIPrefix(o), label=URITail(str(p))))
                            # result[ URITail(str(p))] = trimWellKnownURIPrefix(o)

            # or maybe it's a stats collection
            elif isAStatCollection (nidm_file_tuples, data_object):
                category = 'derivative'
                for (s, p, o) in rdf_graph.triples((data_object, None, None)):
                        cde = getDataTypeInfo(rdf_graph,p )
                        result.append(
                            makeValueTypeFromDataTypeInfo(value=str(o), data_type_info_tuple=cde)
                        )
                        # result[ URITail(str(p)) ] = str(o)

    return ActivityData(category=category, uuid=trimWellKnownURIPrefix(acquisition_uri),  data=result)

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def GetProjectAttributes(nidm_files_tuple, project_id):
    result = {
        ACQUISITION_MODALITY: set([]),
        IMAGE_CONTRAST_TYPE: set([]),
        IMAGE_USAGE_TYPE : set([]),
        TASK : set([])
    }

    project_uuid = expandUUID(project_id)

    for file in nidm_files_tuple:
        rdf_graph = OpenGraph(file)
        #find all the projects
        for (project,pred,o) in rdf_graph.triples((None, None, Constants.NIDM['Project'])):
            #check if it is our project
            if str(project) == str(project_uuid):
                # get all the basic data from the project
                for (proj, predicate, object) in rdf_graph.triples((project, None, None)):
                    result[ matchPrefix(str(predicate)) ] = str(object)

    # now drill into the acquisition objects to get some specific
    # elements: AcquisitionModality, ImageContrastType, ImageUsageType, Task
    sessions = getSessions(nidm_files_tuple, project_id)
    for s in sessions:
        acquistions = getAcquisitions(nidm_files_tuple, s)
        for a in acquistions:
            acq_obj = getActivityData(nidm_files_tuple, a)
            for de in acq_obj.data:
                if de.label == 'hadAcquisitionModality':
                    result[ACQUISITION_MODALITY].add(de.value)
                if de.label == 'hadImageContrastType':
                    result[IMAGE_CONTRAST_TYPE].add(de.value)
                if de.label == 'hadImageUsageType':
                    result[IMAGE_USAGE_TYPE].add(de.value)
                if de.label == 'Task':
                    result[TASK].add(de.value)

    # de-set-ify items so they will play nice with JSON later
    result[ACQUISITION_MODALITY] = list(result[ACQUISITION_MODALITY])
    result[IMAGE_CONTRAST_TYPE] = list(result[IMAGE_CONTRAST_TYPE])
    result[IMAGE_USAGE_TYPE] = list(result[IMAGE_USAGE_TYPE])
    result[TASK] = list(result[TASK])

    return result

@functools.lru_cache(maxsize=BIG_CACHE_SIZE)
def GetAllPredicates(nidm_files_tuple):
    pred_set = set()
    for file in nidm_files_tuple:
        rdf_graph = OpenGraph(file)
        predicates = rdf_graph.predicates()
        for p in predicates:
            pred_set.add(p)
    return pred_set


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def GetDataelements(nidm_files_tuple):
    result = {"data_elements": {"uuid": [], "label": [], "data_type_info": []}}
    found_uris = set()

    for file in nidm_files_tuple:
        rdf_graph = OpenGraph(file)
        #find all the datatypes
        for de_uri in rdf_graph.subjects(predicate=isa, object=Constants.NIDM['DataElement']):
            if de_uri not in found_uris:  # don't add duplicates
                dti = getDataTypeInfo(rdf_graph, de_uri)
                result['data_elements']['uuid'].append(str(dti['dataElementURI']))
                result['data_elements']['label'].append(str(dti['label']))
                result['data_elements']['data_type_info'].append( dti )
                found_uris.add(de_uri)
        # find all the datatypes
        for de_uri in rdf_graph.subjects(predicate=isa, object=Constants.NIDM['PersonalDataElement']):
            if de_uri not in found_uris:  # don't add duplicates
                dti = getDataTypeInfo(rdf_graph, de_uri)
                result['data_elements']['uuid'].append(str(dti['dataElementURI']))
                result['data_elements']['label'].append(str(dti['label']))
                result['data_elements']['data_type_info'].append(dti)
                found_uris.add(de_uri)

    # now look for any of the CDEs
    all_predicates = GetAllPredicates(nidm_files_tuple)
    cde_graph = nidm.experiment.CDE.getCDEs()
    cde_types = cde_graph.subjects(predicate=Constants.RDFS['subClassOf'], object=Constants.NIDM['DataElement'])
    cde_type_set = set() # i.e. fs:DataElement
    known_cde_types = set() # i.e. fs_003579
    for t in cde_types:
        cde_type_set.add(t)
        for s in cde_graph.subjects(predicate=isa, object=t):
            known_cde_types.add(s)


    for predicate in all_predicates:
        if predicate in known_cde_types:
            dti = getDataTypeInfo(cde_graph, predicate)
            result['data_elements']['uuid'].append(str(dti['dataElementURI']))
            result['data_elements']['label'].append(str(dti['label']))
            result['data_elements']['data_type_info'].append(dti)

    return result


def GetDataelementDetails(nidm_files_tuple, dataelement):
    result = {}

    for file in nidm_files_tuple:
        rdf_graph = OpenGraph(file)
        for de_uri in rdf_graph.subjects(predicate=isa, object=Constants.NIDM['DataElement']):
            dti = getDataTypeInfo(rdf_graph, de_uri)

            # check if this is the correct one
            if not (dataelement in [ str(dti['label']), str(dti['dataElement']), str(dti['dataElementURI']) ] ):
                continue

            for key in dti.keys():
                result[key] = dti[key]
            result['inProjects'] = set()

            # figure out what project the dataelement was used in
            uri = dti["dataElementURI"]

            a_list = rdf_graph.subjects(predicate=uri)
            for a in a_list: # a is an assessment / AcquisitionObject
                b_list = rdf_graph.objects(subject=a, predicate=Constants.PROV['wasGeneratedBy'])
                for b in b_list: # b is an Acquisition / Activity
                    c_list = rdf_graph.objects(subject=b, predicate=Constants.DCT['isPartOf'])
                    for c in c_list: # c is a session
                        d_list = rdf_graph.objects(subject=c, predicate=Constants.DCT['isPartOf'])
                        for d in d_list: # d is most likely a project
                            if d in rdf_graph.subjects(predicate=isa, object=Constants.NIDM['Project']):
                                result['inProjects'].add("{} ({})".format(str(d), file))

            return result # found it, we are done


    if result == {}:  # didn't find it yet, check the CDEs
        cde_graph = nidm.experiment.CDE.getCDEs()
        for de_uri in cde_graph.subjects(predicate=isa):
            dti = getDataTypeInfo(cde_graph, de_uri)

            # check if this is the correct one
            if not (dataelement in [str(dti['label']), str(dti['dataElement']), str(dti['dataElementURI'])]):
                continue

            for key in dti.keys():
                result[key] = dti[key]
            result['inProjects'] = set()
            result['inProjects'].add("Common Data Element")

            for file in nidm_files_tuple:
                rdf_graph = OpenGraph(file)
                if result['dataElementURI'] in rdf_graph.predicates():
                    result['inProjects'].add(file)


            return result # found it, we are done

    return result


import nidm.experiment.Navigate
from nidm.experiment import Project, Session, AssessmentAcquisition, AssessmentObject, Acquisition, AcquisitionObject, \
    Query
from nidm.core import Constants
from rdflib import Namespace, URIRef
import prov.model as pm
from os import remove, path, environ
import tempfile
import pytest
from nidm.experiment.CDE import download_cde_files
from nidm.experiment.tools.rest_statistics import GetProjectsComputedMetadata


from prov.model import ProvDocument, QualifiedName
from prov.model import Namespace as provNamespace
import json
import urllib.request
from pathlib import Path

ABIDE_FILES = ('cmu_a.nidm.ttl',)

cmu_test_project_uuid = None
cmu_test_subject_uuid = None


# when set to true, this will test example NIDM files downloaded from
# the GitHub dbkeator/simple2_NIDM_examples repo
#
# DBK: this is a bit unsafe as the TTL files in the github repo above can change and the UUID will change since they are randomly
# generated at this point.  It's probably more robust to explicitly create these files for the time being and explicitly set the
# UUID in the test file:
# For example:  kwargs={Constants.NIDM_PROJECT_NAME:"FBIRN_PhaseIII",Constants.NIDM_PROJECT_IDENTIFIER:1200,Constants.NIDM_PROJECT_DESCRIPTION:"Test investigation2"}
#               project = Project(uuid="_654321",attributes=kwargs)
USE_GITHUB_DATA = True

@pytest.fixture(scope="module", autouse="True")
def setup():
    global cmu_test_project_uuid, cmu_test_subject_uuid

    projects = Query.GetProjectsUUID(ABIDE_FILES)
    for p in projects:
        proj_info = nidm.experiment.Navigate.GetProjectAttributes(ABIDE_FILES, p)
        if 'dctypes:title' in proj_info.keys() and proj_info['dctypes:title'] == 'ABIDE CMU_a Site':
            cmu_test_project_uuid = p
            break
    subjects = Query.GetParticipantIDs(ABIDE_FILES)
    cmu_test_subject_uuid = subjects['uuid'][0]


def test_GetProjectMetadata():

    kwargs={Constants.NIDM_PROJECT_NAME:"FBIRN_PhaseII",Constants.NIDM_PROJECT_IDENTIFIER:9610,Constants.NIDM_PROJECT_DESCRIPTION:"Test investigation"}
    project = Project(uuid="_123456",attributes=kwargs)


    #save a turtle file
    with open("test_gpm.ttl",'w') as f:
        f.write(project.serializeTurtle())

    kwargs={Constants.NIDM_PROJECT_NAME:"FBIRN_PhaseIII",Constants.NIDM_PROJECT_IDENTIFIER:1200,Constants.NIDM_PROJECT_DESCRIPTION:"Test investigation2"}
    project = Project(uuid="_654321",attributes=kwargs)


    #save a turtle file
    with open("test2_gpm.ttl",'w') as f:
        f.write(project.serializeTurtle())


    #WIP test = Query.GetProjectMetadata(["test.ttl", "test2.ttl"])

    #assert URIRef(Constants.NIDM + "_654321") in test
    #assert URIRef(Constants.NIDM + "_123456") in test
    #assert URIRef(Constants.NIDM_PROJECT_IDENTIFIER + "1200") in test
    #assert URIRef(Constants.NIDM_PROJECT_IDENTIFIER + "9610") in test
    #assert URIRef((Constants.NIDM_PROJECT_NAME + "FBIRN_PhaseII")) in test
    #assert URIRef((Constants.NIDM_PROJECT_NAME + "FBIRN_PhaseIII")) in test
    #assert URIRef((Constants.NIDM_PROJECT_DESCRIPTION + "Test investigation")) in test
    #assert URIRef((Constants.NIDM_PROJECT_DESCRIPTION + "Test investigation2")) in test

    remove("test_gpm.ttl")
    remove("test2_gpm.ttl")


def test_GetProjects():

    kwargs={Constants.NIDM_PROJECT_NAME:"FBIRN_PhaseII",Constants.NIDM_PROJECT_IDENTIFIER:9610,Constants.NIDM_PROJECT_DESCRIPTION:"Test investigation"}
    project = Project(uuid="_123456",attributes=kwargs)


    #save a turtle file
    with open("test_gp.ttl",'w') as f:
        f.write(project.serializeTurtle())

    project_list = Query.GetProjectsUUID(["test_gp.ttl"])

    remove("test_gp.ttl")
    assert Constants.NIIRI + "_123456" in [ str(x) for x in project_list]

def test_GetParticipantIDs():

    kwargs={Constants.NIDM_PROJECT_NAME:"FBIRN_PhaseII",Constants.NIDM_PROJECT_IDENTIFIER:9610,Constants.NIDM_PROJECT_DESCRIPTION:"Test investigation"}
    project = Project(uuid="_123456",attributes=kwargs)
    session = Session(uuid="_13579",project=project)
    acq = Acquisition(uuid="_15793",session=session)
    acq2 = Acquisition(uuid="_15795",session=session)

    person=acq.add_person(attributes=({Constants.NIDM_SUBJECTID:"9999"}))
    acq.add_qualified_association(person=person,role=Constants.NIDM_PARTICIPANT)

    person2=acq2.add_person(attributes=({Constants.NIDM_SUBJECTID:"8888"}))
    acq2.add_qualified_association(person=person2,role=Constants.NIDM_PARTICIPANT)

    #save a turtle file
    with open("test_3.ttl",'w') as f:
        f.write(project.serializeTurtle())

    participant_list = Query.GetParticipantIDs(["test_3.ttl"])

    remove("test_3.ttl")
    assert (participant_list['ID'].str.contains('9999').any())
    assert (participant_list['ID'].str.contains('8888').any())

def test_GetProjectInstruments():
    kwargs = {Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII", Constants.NIDM_PROJECT_IDENTIFIER: 9610,
              Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation"}
    proj_uuid = "_123456gpi"
    project = Project(uuid=proj_uuid, attributes=kwargs)

    session = Session(project)
    acq = AssessmentAcquisition(session)

    kwargs={pm.PROV_TYPE:pm.QualifiedName(pm.Namespace("nidm",Constants.NIDM),"NorthAmericanAdultReadingTest")}
    acq_obj = AssessmentObject(acq,attributes=kwargs)

    acq2 = AssessmentAcquisition(session)

    kwargs={pm.PROV_TYPE:pm.QualifiedName(pm.Namespace("nidm",Constants.NIDM),"PositiveAndNegativeSyndromeScale")}
    acq_obj2 = AssessmentObject(acq2,attributes=kwargs)

    #save a turtle file
    with open("test_gpi.ttl",'w') as f:
        f.write(project.serializeTurtle())

    assessment_list = Query.GetProjectInstruments(["test_gpi.ttl"], proj_uuid)

    remove("test_gpi.ttl")

    assert Constants.NIDM + "NorthAmericanAdultReadingTest" in [str(x) for x in assessment_list['assessment_type'].to_list()]
    assert Constants.NIDM + "PositiveAndNegativeSyndromeScale" in [str(x) for x in assessment_list['assessment_type'].to_list()]


'''
The test data file could/should have the following project meta data. Taken from
https://raw.githubusercontent.com/incf-nidash/nidm/master/nidm/nidm-experiment/terms/nidm-experiment.owl
  
  - descrption
  - fileName
  - license
  - source
  - title
  - hadNumericalValue ???
  - BathSolution ???
  - CellType
  - ChannelNumber
  - ElectrodeImpedance
  - GroupLabel
  - HollowElectrodeSolution
  - hadImageContrastType
  - hadImageUsageType
  - NumberOfChannels
  - AppliedFilter
  - SolutionFlowSpeed
  - RecordingLocation
   
Returns the 
  
'''
def saveTestFile(file_name, data):
    project = Project(uuid="_123_" + file_name, attributes=data)

    return saveProject(file_name, project)

def saveProject(file_name, project):
    # save a turtle file
    with open(file_name, 'w') as f:
        f.write(project.serializeTurtle())
    return "nidm:_123_{}".format(file_name)


def makeProjectTestFile(filename):
    DCTYPES = Namespace("http://purl.org/dc/dcmitype/")

    kwargs = {Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII", # this is the "title"
              Constants.NIDM_PROJECT_IDENTIFIER: 9610,
              Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
              Constants.NIDM_FILENAME: "testfile.ttl",
              Constants.NIDM_PROJECT_LICENSE: "MIT Licence",
              Constants.NIDM_PROJECT_SOURCE: "Educational Source",
              Constants.NIDM_HAD_NUMERICAL_VALUE: "numval???",
              Constants.NIDM_BATH_SOLUTION: "bath",
              Constants.NIDM_CELL_TYPE: "ctype",
              Constants.NIDM_CHANNEL_NUMBER: "5",
              Constants.NIDM_ELECTRODE_IMPEDANCE: ".01",
              Constants.NIDM_GROUP_LABEL: "group 123",
              Constants.NIDM_HOLLOW_ELECTRODE_SOLUTION: "water",
              Constants.NIDM_HAD_IMAGE_CONTRACT_TYPE: "off",
              Constants.NIDM_HAD_IMAGE_USAGE_TYPE: "abcd",
              Constants.NIDM_NUBMER_OF_CHANNELS: "11",
              Constants.NIDM_APPLIED_FILTER: "on",
              Constants.NIDM_SOLUTION_FLOW_SPEED: "2.8",
              Constants.NIDM_RECORDING_LOCATION: "lab"
              }
    return saveTestFile(filename, kwargs)

def makeProjectTestFile2(filename):
    DCTYPES = Namespace("http://purl.org/dc/dcmitype/")

    kwargs = {Constants.NIDM_PROJECT_NAME: "TEST B", # this is the "title"
              Constants.NIDM_PROJECT_IDENTIFIER: 1234,
              Constants.NIDM_PROJECT_DESCRIPTION: "More Scans",
              Constants.NIDM_FILENAME: "testfile2.ttl",
              Constants.NIDM_PROJECT_LICENSE: "Creative Commons",
              Constants.NIDM_PROJECT_SOURCE: "Other",
              Constants.NIDM_HAD_NUMERICAL_VALUE: "numval???",
              Constants.NIDM_BATH_SOLUTION: "bath",
              Constants.NIDM_CELL_TYPE: "ctype",
              Constants.NIDM_CHANNEL_NUMBER: "5",
              Constants.NIDM_ELECTRODE_IMPEDANCE: ".01",
              Constants.NIDM_GROUP_LABEL: "group 123",
              Constants.NIDM_HOLLOW_ELECTRODE_SOLUTION: "water",
              Constants.NIDM_HAD_IMAGE_CONTRACT_TYPE: "off",
              Constants.NIDM_HAD_IMAGE_USAGE_TYPE: "abcd",
              Constants.NIDM_NUBMER_OF_CHANNELS: "11",
              Constants.NIDM_APPLIED_FILTER: "on",
              Constants.NIDM_SOLUTION_FLOW_SPEED: "2.8",
              Constants.NIDM_RECORDING_LOCATION: "lab"
              }
    project = Project(uuid="_123_" + filename, attributes=kwargs)
    s1 = Session(project)

    a1 = AssessmentAcquisition(session=s1)
      # = s1.add_acquisition("a1", attributes={"http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#Age" : 22})

    p1 = a1.add_person("p1", attributes={Constants.NIDM_GIVEN_NAME:"George", Constants.NIDM_AGE: 22})
    a1.add_qualified_association(person=p1, role=Constants.NIDM_PARTICIPANT)


    return saveProject(filename, project)


def test_GetProjectsMetadata():
    p1 = makeProjectTestFile("testfile.ttl")
    p2 = makeProjectTestFile2("testfile2.ttl")
    files = ["testfile.ttl", "testfile2.ttl"]

    if USE_GITHUB_DATA and not Path('./cmu_a.nidm.ttl').is_file():
        urllib.request.urlretrieve(
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
            "cmu_a.nidm.ttl"
        )
        files.append("cmu_a.nidm.ttl")
    elif Path('./cmu_a.nidm.ttl').is_file():
        files.append("cmu_a.nidm.ttl")

    parsed = Query.GetProjectsMetadata(files)

    # assert parsed['projects'][p1][str(Constants.NIDM_PROJECT_DESCRIPTION)] == "Test investigation"
    # assert parsed['projects'][p2][str(Constants.NIDM_PROJECT_DESCRIPTION)] == "More Scans"

    # we shouldn't have the computed metadata in this result
    # assert parsed['projects'][p1].get (Query.matchPrefix(str(Constants.NIDM_NUMBER_OF_SUBJECTS)), -1) == -1

    if USE_GITHUB_DATA:
        # find the project ID from the CMU file
        p3 = None
        for project_id in parsed['projects']:
            if project_id != p1 and project_id != p2:
                if parsed['projects'][project_id][str(Constants.NIDM_PROJECT_NAME)] == "ABIDE CMU_a Site":
                    p3 = project_id
                    break
        assert p3 != None


#
# moved to test_rest.py
#
# def test_GetProjectsComputedMetadata():
#
#     p1 = makeProjectTestFile("testfile.ttl")
#     p2 = makeProjectTestFile2("testfile2.ttl")
#     files = ["testfile.ttl", "testfile2.ttl"]
#
#     if USE_GITHUB_DATA:
#         if not Path('./cmu_a.nidm.ttl').is_file():
#             urllib.request.urlretrieve (
#                 "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
#                 "cmu_a.nidm.ttl"
#             )
#         files.append("cmu_a.nidm.ttl")
#
#     parsed = Query.GetProjectsComputedMetadata(files)
#
#     if USE_GITHUB_DATA:
#         for project_id in parsed['projects']:
#             if project_id != p1 and project_id != p2:
#                 p3 = project_id
#         assert parsed['projects'][p3][str(Constants.NIDM_PROJECT_NAME)] == "ABIDE CMU_a Site"
#         assert parsed['projects'][p3][Query.matchPrefix(str(Constants.NIDM_NUMBER_OF_SUBJECTS))] == 14
#         assert parsed['projects'][p3]["age_min"] == 21.0
#         assert parsed['projects'][p3]["age_max"] == 33.0
#         assert parsed['projects'][p3][str(Constants.NIDM_GENDER)] == ['1', '2']


def test_prefix_helpers():

    assert Query.expandNIDMAbbreviation("ndar:src_subject_id") == "https://ndar.nih.gov/api/datadictionary/v2/dataelement/src_subject_id"

    assert Query.matchPrefix("http://purl.org/nidash/nidm#abc") == "nidm:abc"
    assert Query.matchPrefix("http://www.w3.org/ns/prov#123") == "prov:123"
    assert Query.matchPrefix("http://purl.org/nidash/fsl#xyz") == "fsl:xyz"
    assert Query.matchPrefix("http://purl.org/nidash/fsl#xyz", short=True) == "fsl"


def test_getProjectAcquisitionObjects():
    if not Path('./cmu_a.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
            "cmu_a.nidm.ttl"
        )
    files = ['cmu_a.nidm.ttl']

    project_list = Query.GetProjectsUUID(files)
    project_uuid = str(project_list[0])
    objects = Query.getProjectAcquisitionObjects(files,project_uuid)

    assert isinstance(objects,list)


def test_GetProjectAttributes():
    global cmu_test_project_uuid
    if not Path('./cmu_a.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
            "cmu_a.nidm.ttl"
        )
    files = ABIDE_FILES

    project_uuid = cmu_test_project_uuid
    project_attributes = nidm.experiment.Navigate.GetProjectAttributes(files, project_uuid)
    assert ('prov:Location' in project_attributes) or ('Location' in project_attributes)
    assert ('dctypes:title' in project_attributes) or ('title' in project_attributes)
    assert ('http://www.w3.org/1999/02/22-rdf-syntax-ns#type' in project_attributes) or ('type' in project_attributes)
    assert ('AcquisitionModality') in project_attributes
    assert ('ImageContrastType') in project_attributes
    assert ('Task') in project_attributes
    assert ('ImageUsageType') in project_attributes


def test_download_cde_files():
    cde_dir = download_cde_files()
    assert cde_dir == tempfile.gettempdir()
    fcount = 0
    for url in Constants.CDE_FILE_LOCATIONS:
        fname = url.split('/')[-1]
        assert path.isfile("{}/{}".format(cde_dir, fname) )
        fcount += 1
    assert fcount > 0

@pytest.mark.skip(reason="We don't have an easily accessible file for this test so skipping it until better test samples are available.")
def test_custom_data_types():
    SPECIAL_TEST_FILES = ['/opt/project/ttl/MTdemog_aseg.ttl']

    valuetype1 = Query.getDataTypeInfo(Query.OpenGraph(SPECIAL_TEST_FILES[0]), 'no-real-value')
    assert (valuetype1 == False)

    valuetype2 = Query.getDataTypeInfo(Query.OpenGraph(SPECIAL_TEST_FILES[0]), Constants.NIIRI['age_e3hrcc'])
    assert (str(valuetype2['label']) == 'age')
    assert (str(valuetype2['description']) == "Age of participant at scan")
    assert (str(valuetype2['isAbout']) == str(Constants.NIIRI['24d78sq']))

    valuetype3 = Query.getDataTypeInfo(Query.OpenGraph(SPECIAL_TEST_FILES[0]), 'age_e3hrcc')
    assert (str(valuetype3['label']) == 'age')
    assert (str(valuetype3['description']) == "Age of participant at scan")
    assert (str(valuetype3['isAbout']) == str(Constants.NIIRI['24d78sq']))
